""" Basic description of a sweep in DMT.

Sweeps are the basic element that can be fed into simulators or be retrieved from simulations or measurements.
Features:

* Clear syntax and definition to create a well described simulation independent of the simulator interface.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from __future__ import annotations
from typing import Dict, List, Mapping, Type, Optional, Union, Tuple
import numpy as np
from DMT.core import (
    get_specifier_from_string,
    SpecifierStr,
)


class SweepDef(object):
    """One sweep definition

    Subclass this to introduce own sweep types. The methods which need to be overwritten are:

    * :meth:`~DMT.core.sweep.SweepDef._correct_sweep_type`
    * :meth:`~DMT.core.sweep.SweepDef.set_values`

    It is recommended try the super method and except the raised errors.

    Parameters
    ----------
    var_name : :class:`~DMT.core.naming.SpecifierStr`
        Name of the variable to sweep.
    sweep_type : str
        Which type of sweep is performed for this variable
    sweep_order : int, optional
        Position in the sweep.
    value_def : List or np.array, optional
        The needed variables depending on the sweep type.
    sync : :class:`~DMT.core.naming.SpecifierStr` or :class:`~DMT.core.sweep.SweepDef`, optional
        Synced 'slave' sweep.
    """

    def __init__(
        self,
        var_name: SpecifierStr,
        sweep_type: str,
        sweep_order: Optional[int] = None,
        value_def: Optional[Union[List, np.array]] = None,
        master: Optional[SpecifierStr] = None,
        offset: Optional[Union[int, float, SpecifierStr]] = None,
        sync: Optional[Union[SpecifierStr, SweepDef]] = None,
    ):
        # needed names for repr
        self._attr_repr = [
            "var_name",
            "sweep_type",
            "sweep_order",
            "value_def",
            "master_var",
            "sync_var",
            "offset_var",
            "offset_value",
        ]

        self.var_name = get_specifier_from_string(var_name)
        if not isinstance(self.var_name, SpecifierStr):
            raise IOError(
                "DMT->Sweep: Automatic conversion from var_name string to SpecifierStr failed for the variable: "
                + var_name
                + "!\n Please provide the correct SpecifierStr manually. "
            )

        self._sweep_type = None
        self.sweep_type = sweep_type

        self.sweep_order = sweep_order

        # check if either value_def or master is given-----------
        if value_def is None and master is None:
            raise IOError(
                "DMT -> Sweep: You either forgot value_def or master in your sweepdef for the variable "
                + self.var_name
                + "."
            )

        self._master_var = None
        self._master_swd = None
        self.master = master

        self._sync_var = None
        self._sync_swd = None
        self.sync = sync

        if value_def is None:
            self.value_def = np.zeros((1,))
        else:
            self.value_def = value_def

        self._offset_var = None
        self._offset_swd = None
        self._offset_value = None
        self.offset = offset

        self.values = None
        self.set_values()

    @property
    def sweep_type(self):
        """Directly return the sweep type"""
        return self._sweep_type

    @sweep_type.setter
    def sweep_type(self, sweep_type_new):
        """Call the corrector. It can be overwritten by a inheriting class."""
        self._sweep_type = self._correct_sweep_type(sweep_type_new)

    def _correct_sweep_type(self, sweep_type_new):
        """Correct the sweep type string to allowed ones.

        Parameters
        ----------
        sweep_type_new : str

        Returns
        -------
        sweep_type : str

        Raises
        ------
        IOError
            If the sweep type is not known.
        """
        sweep_type_new = sweep_type_new.upper()
        # correct the sweep type
        if "LIN" in sweep_type_new:
            return "LIN"
        elif "LOG" in sweep_type_new:
            return "LOG"
        elif "CON" in sweep_type_new:
            return "CON"
        elif "SYNC" in sweep_type_new:
            return "SYNC"
        elif "LIST" in sweep_type_new:
            return "LIST"
        else:
            raise IOError(
                'DMT->Sweep: specified sweeptype:"' + sweep_type_new + '" is unknown to DMT.'
            )

    @property
    def master(self):
        """Master sweep definition, None if the master sweep is not set or not given."""
        return self._master_swd

    @property
    def master_var(self):
        """Getter of the master_var. This attribute has no setter! Use master itself for this!"""
        return self._master_var

    @master.setter
    def master(self, master_new):
        """Sets the master sweep and at the same time the master_var attribute to keep them consistant."""
        try:
            self._master_var = master_new.var_name
            self._master_swd = master_new
        except AttributeError:
            self._master_var = master_new
            self._master_swd = None

    @property
    def sync(self):
        """Sync sweep definition, None if there is no synced sweeped."""
        return self._sync_swd

    @property
    def sync_var(self):
        """Getter of the sync_var. This attribute has no setter! Use sync itself for this!"""
        return self._sync_swd

    @sync.setter
    def sync(self, sync_new):
        """Sets the sync sweep and at the same time the sync_var attribute to keep them consistant."""
        try:
            self._sync_var = sync_new.var_name
            self._sync_swd = sync_new
        except AttributeError:
            self._sync_var = sync_new
            self._sync_swd = None

    @property
    def offset(self):
        """Offset sweep definition. Can be None, float64 or a SweepDef"""
        return self._offset_swd

    @property
    def offset_var(self):
        """Getter of the offset_var. This attribute has no setter! Use offset itself for this! None, if offset is not a SweepDef."""
        return self._offset_var

    @property
    def offset_value(self):
        """Getter of the offset_valaue. This attribute has no setter! Use offset itself for this! None, if offset is a SweepDef."""
        return self._offset_value

    @offset.setter
    def offset(self, offset_new):
        """Sets the offset sweep and at the same time the offset_var attribute to keep them consistant."""
        self._offset_value = None
        try:
            self._offset_var = offset_new.var_name
            self._offset_swd = offset_new
        except AttributeError:
            if isinstance(offset_new, str):
                self._offset_var = offset_new
                self._offset_swd = None
            else:
                self._offset_var = None
                self._offset_value = offset_new
                self._offset_swd = offset_new

    def __repr__(self):
        return (
            "SweepDef ("
            + ", ".join(
                ["'{0:s}': {1}".format(attr, getattr(self, attr)) for attr in self._attr_repr]
            )
            + ")"
        )

    def set_values(self):
        """Sets self.values

        Raises
        ------
        OSError
            If the sweep type is not known.
        """
        if self.sweep_type == "LIN":
            self.values = np.linspace(
                self.value_def[0], self.value_def[1], self.value_def[2], dtype=np.float64
            )

        elif self.sweep_type == "CON":
            self.values = np.array(self.value_def, dtype=np.float64)
            if self.values.size > 1:
                raise IOError(
                    "DMT->SweepDef: Constant sweeps must have only one value in value_def!"
                )

        elif self.sweep_type == "LOG":
            self.values = np.logspace(
                self.value_def[0], self.value_def[1], self.value_def[2], dtype=np.float64
            )

        elif self.sweep_type == "LIST":
            self.values = np.array(self.value_def, dtype=np.float64)

        elif self.sweep_type == "SYNC":
            if self.master is None:  # master has not been replaced jet..
                return

            try:
                self.values = self.master.values + self.offset
            except TypeError:
                self.values = np.zeros((self.offset.values.size, self.master.values.size))
                for i_col in range(self.values.shape[0]):
                    self.values[i_col, :] = self.master.values + self.offset.values[i_col]
                self.values = np.concatenate(self.values)


class SweepDefConst(SweepDef):
    """Constant sweep definition

    Parameters
    ----------
    var_name : SpecifierStr
        Variable name
    value_def : Union[int, float]
        Value for the variable
    sweep_order : int, optional
        Position in the sweep.
    """

    def __init__(
        self,
        var_name: SpecifierStr,
        value_def: Union[int, float, List, np.array],
        sweep_order: Optional[int] = None,
    ):
        if isinstance(value_def, (int, float)):
            value_def = [value_def]

        super().__init__(
            var_name=var_name,
            sweep_type="CON",
            sweep_order=sweep_order,
            value_def=np.asarray(value_def),
        )


class SweepDefLinear(SweepDef):
    """Linear sweep definition

    Parameters
    ----------
    var_name : SpecifierStr
        Variable name
    value_def : Union[int, float], optional
        Values for the linear sweepdef. Must have 3 Elements: np.linspace(self.value_def[0], self.value_def[1], self.value_def[2], dtype=np.float64)
    start : [float, int], optional
        Start value for np.linspace (always included)
    stop : float, int], optional
        Stop value for np.linspace (always included)
    steps : int, optional
        Number of steps for np.linspace

    sweep_order : int, optional
        Position in the sweep.
    """

    def __init__(
        self,
        var_name: SpecifierStr,
        value_def: Optional[Union[Tuple, List, np.array]] = None,
        start: Optional[Union[float, int]] = None,
        stop: Optional[Union[float, int]] = None,
        steps: Optional[int] = None,
        sweep_order: Optional[int] = None,
    ):
        if value_def is None:
            value_def = np.array([start, stop, steps])
        super().__init__(
            var_name=var_name,
            sweep_type="LIN",
            sweep_order=sweep_order,
            value_def=value_def,
        )


class SweepDefLog(SweepDef):
    """Logarithmic sweep definition

    Parameters
    ----------
    var_name : SpecifierStr
        Variable name
    value_def : Union[int, float], optional
        Values for the logarithmic sweepdef. Must have 3 Elements: np.logspace(self.value_def[0], self.value_def[1], self.value_def[2], dtype=np.float64)
    start : [float, int], optional
        Start value for np.logspace (always included)
    stop : float, int], optional
        Stop value for np.logspace (always included)
    steps : int, optional
        Number of steps for np.logspace

    sweep_order : int, optional
        Position in the sweep.
    """

    def __init__(
        self,
        var_name: SpecifierStr,
        value_def: Optional[Union[Tuple, List, np.array]] = None,
        start: Optional[Union[float, int]] = None,
        stop: Optional[Union[float, int]] = None,
        steps: Optional[int] = None,
        sweep_order: Optional[int] = None,
    ):
        if value_def is None:
            value_def = np.array([start, stop, steps])
        super().__init__(
            var_name=var_name,
            sweep_type="LOG",
            sweep_order=sweep_order,
            value_def=value_def,
        )


class SweepDefList(SweepDef):
    """List or table sweep definition

    Parameters
    ----------
    var_name : SpecifierStr
        Variable name
    value_def : Union[int, float]
        Values for the variable.

    sweep_order : int, optional
        Position in the sweep.
    """

    def __init__(
        self,
        var_name: SpecifierStr,
        value_def: Union[Tuple, List, np.array],
        sweep_order: Optional[int] = None,
    ):
        super().__init__(
            var_name=var_name,
            sweep_type="LIST",
            sweep_order=sweep_order,
            value_def=value_def,
        )


class SweepDefSync(SweepDef):
    """Synchronized sweep definition

    Parameters
    ----------
    var_name : SpecifierStr
        Variable name
    value_def : Union[int, float]
        Values for the variable.
    master : SpecifierStr
        Master sweep variable.
    offset : Union[int, float, SpecifierStr]
        Offset value for the sweeps or a variable name, which is sweeped in a different SweepDef.

    sweep_order : int, optional
        Position in the sweep.
    """

    def __init__(
        self,
        var_name: SpecifierStr,
        master: SpecifierStr,
        offset: Union[int, float, SpecifierStr],
        sweep_order: Optional[int] = None,
    ):
        super().__init__(
            var_name=var_name,
            sweep_type="SYNC",
            sweep_order=sweep_order,
            master=master,
            offset=offset,
        )



""" test ADS input file generation, copy to server, run simulation and copy back

"""
import types
import logging
import numpy as np
from pathlib import Path

from DMT.core import DutType, Sweep, specifiers, SimCon, Plot, MCard, VAFile
from DMT.core.circuit import (
    Circuit,
    CircuitElement,
    RESISTANCE,
    CAPACITANCE,
    SHORT,
    VOLTAGE,
    SGP_BJT,
)
from DMT.xyce import DutXyce


folder_path = Path(__file__).resolve().parent
logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_xyce_sgp.log",
    filemode="w",
)


def get_circuit(self, use_build_in=False, **kwargs):
    """Returns a circuit which uses the modelcard to which the method is attached.

    Parameter
    ------------
    use_build_in : bool
        Choose which model to use.

    Returns
    -------
    circuit : :class:`~DMT.core.circuit.Circuit`

    """
    if use_build_in:
        module_name = SGP_BJT
    else:
        module_name = self.default_module_name  # should be "bjtn" because vae

    node_emitter = next(f"n_{node.upper()}" for node in self.nodes_list if "E" in node.upper())
    node_base = next(f"n_{node.upper()}" for node in self.nodes_list if "B" in node.upper())
    node_collector = next(f"n_{node.upper()}" for node in self.nodes_list if "C" in node.upper())

    circuit_elements = []
    # model instance
    circuit_elements.append(
        CircuitElement(
            module_name,
            "Q_H",
            [f"n_{node.upper()}" for node in self.nodes_list],
            # ["n_C", "n_B", "n_E"],
            parameters=self,
        )
    )

    # BASE NODE CONNECTION #############
    # metal resistance between contact base point and real collector
    try:
        rbm = self.get("_rbm").value
    except KeyError:
        rbm = 1e-3

    circuit_elements.append(
        CircuitElement(RESISTANCE, "Rbm", ["n_B_FORCED", node_base], parameters=[("R", str(rbm))])
    )
    # shorts for current measurement
    circuit_elements.append(
        CircuitElement(
            SHORT,
            "I_B",
            ["n_BX", "n_B_FORCED"],
        )
    )
    # capacitance since AC already deembeded Rbm
    circuit_elements.append(
        CircuitElement(CAPACITANCE, "Cbm", ["n_B_FORCED", node_base], parameters=[("C", str(1))])
    )

    # COLLECTOR NODE CONNECTION #############
    circuit_elements.append(
        CircuitElement(
            SHORT,
            "I_C",
            ["n_CX", "n_C_FORCED"],
        )
    )
    # metal resistance between contact collector point and real collector
    try:
        rcm = self.get("_rcm").value
    except KeyError:
        rcm = 1e-3

    circuit_elements.append(
        CircuitElement(
            RESISTANCE, "Rcm", ["n_C_FORCED", node_collector], parameters=[("R", str(rcm))]
        )
    )
    # capacitance since AC already deembeded Rcm
    circuit_elements.append(
        CircuitElement(
            CAPACITANCE, "Ccm", ["n_C_FORCED", node_collector], parameters=[("C", str(1))]
        )
    )
    # EMITTER NODE CONNECTION #############
    circuit_elements.append(
        CircuitElement(
            SHORT,
            "I_E",
            ["n_EX", "n_E_FORCED"],
        )
    )
    # metal resistance between contact emiter point and real emiter
    try:
        rem = self.get("_rem").value
    except KeyError:
        rem = 1e-3

    circuit_elements.append(
        CircuitElement(
            RESISTANCE, "Rem", ["n_E_FORCED", node_emitter], parameters=[("R", str(rem))]
        )
    )
    # capacitance since AC already deembeded Rcm
    circuit_elements.append(
        CircuitElement(CAPACITANCE, "Cem", ["n_E_FORCED", node_emitter], parameters=[("C", str(1))])
    )
    # add sources and thermal resistance
    circuit_elements.append(
        CircuitElement(
            VOLTAGE, "V_B", ["n_BX", "0"], parameters=[("Vdc", "V_B"), ("Vac", "V_B_ac")]
        )
    )
    circuit_elements.append(
        CircuitElement(
            VOLTAGE, "V_C", ["n_CX", "0"], parameters=[("Vdc", "V_C"), ("Vac", "V_C_ac")]
        )
    )
    circuit_elements.append(
        CircuitElement(
            VOLTAGE, "V_E", ["n_EX", "0"], parameters=[("Vdc", "V_E"), ("Vac", "V_E_ac")]
        )
    )

    # metal resistance between contact emitter potential and substrate contact
    if len(self.nodes_list) > 3:
        node_substrate = next(
            f"n_{node.upper()}" for node in self.nodes_list if "S" in node.upper()
        )

        circuit_elements.append(
            CircuitElement(
                SHORT,
                "I_S",
                ["n_SX", node_substrate],
            )
        )
        try:
            rsm = self.get("_rsm").value
        except KeyError:
            rsm = 5
        circuit_elements.append(
            CircuitElement(RESISTANCE, "R_S", ["n_SX", "n_EX"], parameters=[("R", str(rsm))])
        )
    circuit_elements += [
        "V_B=0",
        "V_C=0",
        "V_E=0",
        "ac_switch=0",
        "V_B_ac=1-ac_switch",
        "V_C_ac=ac_switch",
        "V_E_ac=0",
    ]

    return Circuit(circuit_elements)


def get_dut_build_in():
    modelcard = MCard(
        ["c", "b", "e", "s"],
        "QSGP1",
        SGP_BJT,
        1.0,
        va_file=folder_path / "sgp_v1p0.va",
    )
    modelcard.default_module_name = SGP_BJT
    modelcard.get_circuit = types.MethodType(get_circuit, modelcard)
    modelcard.load_model_parameters(folder_path / "bjt.lib")
    return DutXyce(
        None,
        DutType.npn,
        modelcard,
        name="Xyce_BI_",
        nodes="C,B,E",
        copy_va_files=True,
        reference_node="E",
        get_circuit_arguments={"use_build_in": True},
    )


def get_dut_va():
    modelcard = MCard(
        ["c", "b", "e", "s"],
        "QSGP2",
        SGP_BJT,
        1.0,
        va_file=folder_path / "sgp_v1p0.va",
    )
    modelcard.get_circuit = types.MethodType(get_circuit, modelcard)
    modelcard.load_model_parameters(folder_path / "bjt.lib")
    return DutXyce(
        None,
        DutType.npn,
        modelcard,
        name="Xyce_VA_",
        nodes="C,B,E,S,T",
        copy_va_files=True,
        reference_node="E",
        get_circuit_arguments={"use_build_in": False},
    )


def get_sweep():
    sweepdef = [
        {"var_name": "FREQ", "sweep_order": 4, "sweep_type": "LOG", "value_def": [6, 7, 11]},
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0, 1, 11],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 3,
            "sweep_type": "SYNC",
            "master": specifiers.VOLTAGE + "B",
            "offset": specifiers.VOLTAGE + ["C", "B"],
        },
        {
            "var_name": specifiers.VOLTAGE + ["C", "B"],
            "sweep_order": 2,
            "sweep_type": "LIST",
            # "value_def": [0.3, 0, -0.3],
            "value_def": [0],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    return Sweep(
        "gummel",
        sweepdef=sweepdef,
        outputdef=[specifiers.CURRENT + "C", specifiers.CURRENT + "B"],
        othervar={"TEMP": 300},
    )


def test_run_and_read():
    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_ib = specifiers.CURRENT + "B"
    col_ic = specifiers.CURRENT + "C"

    duts_test = []
    duts_test.append(get_dut_build_in())
    duts_test.append(get_dut_va())
    sweep_test = get_sweep()

    sim_con = SimCon(t_max=80)  # damn VA-Compile time
    sim_con.append_simulation(dut=duts_test, sweep=sweep_test)
    sim_con.run_and_read(force=True)

    for dut in duts_test:
        dut_name = dut.name[5:7]
        df = dut.get_data(sweep=sweep_test)

        df.ensure_specifier_column(col_vbe, ports=dut.nodes)
        df.ensure_specifier_column(col_vbc, ports=dut.nodes)

        for _index, vbc, data in df.iter_unique_col(col_vbc, decimals=3):
            data = data[data[specifiers.FREQUENCY] == 1e7]

            if "BI" in dut_name:
                vbe_test = np.array(
                    [
                        5.58161701e-41,
                        1.00000000e-01,
                        2.00000000e-01,
                        3.00000000e-01,
                        4.00000000e-01,
                        5.00000000e-01,
                        6.00000000e-01,
                        6.99999997e-01,
                        7.99999882e-01,
                        8.99996453e-01,
                        9.99979245e-01,
                    ]
                )
                ic_test = np.array(
                    [
                        -1.46936794e-39,
                        9.72598538e-14,
                        2.00083790e-13,
                        8.32249178e-13,
                        2.62352010e-11,
                        1.19704585e-09,
                        5.52532891e-08,
                        2.54706085e-06,
                        1.15610970e-04,
                        3.46916517e-03,
                        2.00870352e-02,
                    ]
                )
                ib_test = np.array(
                    [
                        -2.71176588e-38,
                        5.56179733e-15,
                        3.10153077e-14,
                        1.62060498e-13,
                        1.02019426e-12,
                        1.45616330e-11,
                        5.22082372e-10,
                        2.40877044e-08,
                        1.13533053e-06,
                        3.87381699e-05,
                        3.34227460e-04,
                    ]
                )
            elif "VA" in dut_name:
                vbe_test = np.array(
                    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.79999988, 0.89999645, 0.99997925]
                )
                ic_test = np.array(
                    [
                        0.00000000e00,
                        1.00254762e-13,
                        2.12063425e-13,
                        8.59203354e-13,
                        2.62831194e-11,
                        1.19712072e-09,
                        5.52533967e-08,
                        2.54706098e-06,
                        1.15610969e-04,
                        3.46915636e-03,
                        2.00867109e-02,
                    ]
                )
                ib_test = np.array(
                    [
                        0.00000000e00,
                        1.04817715e-13,
                        2.29527142e-13,
                        4.59828250e-13,
                        1.41721793e-12,
                        1.50579126e-11,
                        5.22677906e-10,
                        2.40883991e-08,
                        1.13533131e-06,
                        3.87380622e-05,
                        3.34219784e-04,
                    ]
                )
            else:
                raise NotImplementedError

            assert np.isclose(np.real(data[col_vbe]), vbe_test).all()
            assert np.isclose(np.real(data[col_ic]), ic_test).all()
            assert np.isclose(np.real(data[col_ib]), ib_test).all()

    return duts_test, sweep_test


if __name__ == "__main__":
    duts, sweep = test_run_and_read()

    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_ib = specifiers.CURRENT + "B"
    col_ic = specifiers.CURRENT + "C"

    plt_gummel = Plot(
        r"$I_{\mathrm{C}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{C}}$",
        y_log=True,
        style="bw",
    )
    plt_ib = Plot(
        r"$I_{\mathrm{B}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{B}}$",
        y_log=True,
        style="bw",
    )

    for dut in duts:
        dut_name = dut.name[5:7]
        df = dut.get_data(sweep=sweep)

        df.ensure_specifier_column(col_vbe, ports=dut.nodes)
        df.ensure_specifier_column(col_vbc, ports=dut.nodes)

        for _index, vbc, data in df.iter_unique_col(col_vbc, decimals=3):
            data = data[data[specifiers.FREQUENCY] == 1e7]

            plt_gummel.add_data_set(
                data[col_vbe],
                data[col_ic],
                label=dut_name + " $V_{{BC}} = {0:1.2f} V$".format(data[col_vbc].to_numpy()[0]),
            )
            plt_ib.add_data_set(
                data[col_vbe],
                data[col_ib],
                label=dut_name + " $V_{{BC}} = {0:1.2f} V$".format(data[col_vbc].to_numpy()[0]),
            )

    plt_ib.plot_pyqtgraph(show=False)
    plt_gummel.plot_pyqtgraph(show=True)



""" Testing the class DMT.core.DutType """

import logging
from pathlib import Path

# import pytest
from DMT.core.dut_type import DutType

folder_path = Path(__file__).resolve().parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_dut_type.log",
    filemode="w",
)


def test_subtypes():
    """Test if the types are the correct subtypes."""

    assert DutType.device & DutType.transistor
    assert DutType.device & DutType.npn
    assert not DutType.device & DutType.tlm  # a tlm is no device
    assert not DutType.meas_struct & DutType.npn

    assert not DutType.cap & DutType.meas_struct
    assert DutType.cap_ac & DutType.meas_struct

    assert DutType.tlmb & DutType.pn_diode  # !!! WRONG!!
    assert not DutType.tlmb.is_subtype(
        DutType.pn_diode
    )  # USE THIS!! To test for flags and not for special devices!


def test_get_nodes():
    """Tests the get nodes of the flags."""

    assert DutType.tetrode.get_nodes() == ["B1", "B2", "E", "C", "S"]


def test_to_string():
    npn = DutType.deem_open_bjt
    assert str(npn) == "open"


if __name__ == "__main__":
    test_subtypes()
    test_get_nodes()
    test_to_string()
    dummy = 1



""" Testing the DMT core utilities
"""
from DMT.core import flatten, enumerate_reversed, strictly_increasing
from DMT.external import resolve_siunitx, tex_to_text

import numpy as np


def test_strictly_increasing():
    assert strictly_increasing(np.array([0.0, 1.0, 2.0, 3.0]))
    assert not strictly_increasing(np.array([0.0, 2.0, 1.0, 3.0]))
    assert not strictly_increasing(np.array([0.0, 1.0, 1.0, 3.0]))


def test_flatten():
    assert (0, 1, 2, 3) == tuple(flatten((0, (1, 2), 3)))
    assert (0, 1, 2, 3) == tuple(
        flatten(
            (
                0,
                (
                    1,
                    (
                        2,
                        3,
                    ),
                ),
            )
        )
    )


def test_enumerate_reversed():
    test_list = ["A", "B", "C", "D"]

    old_index = len(test_list)
    for index, value in enumerate_reversed(test_list):
        assert test_list[index] == value  # original index is preserved

        assert index == old_index - 1  # iterating reversed
        old_index = index

    old_index = 2
    for index, value in enumerate_reversed(test_list, stop=2):
        assert test_list[index] == value  # original index is preserved

        assert index == old_index - 1  # iterating reversed
        old_index = index


def test_resolve_siunitx():
    # stay the same if no si is mentioned
    assert resolve_siunitx("pK") == "pK"

    assert resolve_siunitx("\\si{\\pico\\kelvin}") == "pK"
    assert resolve_siunitx("\\si{\\pico\\kelvin}\\,\\si{\\milli\\volt}") == "pK mV"
    assert resolve_siunitx("\\SI{5}{\\pico\\kelvin}") == "5pK"
    assert resolve_siunitx("\\SI{5}{\\pico\\kelvin}\\,\\SI{0.5}{\\milli\\volt}") == "5pK 0.5mV"

    assert (
        resolve_siunitx(
            "\\underline{Y}^{s}_{\\mathrm{21}} = \\SI{1e5}{\\micro\\siemens\\per\\square\\micro\\meter}"
        )
        == "Y^{s}_{\\mathrm{21}} = 1e5uSinvsqum"
    )


def test_tex_to_text():
    assert tex_to_text(r"I_{C}") == r"I_C"
    assert tex_to_text(r"I_{\mathrm{C}}") == r"I_C"
    assert tex_to_text(r"I_{C}") == r"I_C"


if __name__ == "__main__":
    test_strictly_increasing()
    test_flatten()
    test_enumerate_reversed()
    test_resolve_siunitx()
    test_tex_to_text()



"""
This example shows the basic data handling with python.
We will read data, clean it, calculate with the data and then plot.
"""
from pathlib import Path
from DMT.core import read_data, specifiers, Plot

# path to static folder:
path_static = Path(__file__).resolve().parent.parent / "_static"

# some specifiers, they allow consistent access to electrical data
# all interfaces of DMT adhere to these specifiers!
col_vbe = specifiers.VOLTAGE + ["B", "E"]
col_vbc = specifiers.VOLTAGE + ["B", "C"]
col_ic = specifiers.CURRENT + "C"
col_ib = specifiers.CURRENT + "B"

# read data using the given method
data = read_data(path_static / "HBT_vbc.elpa")

# "clean" the data so we have DMT-specifier in it.
# This ensure that the read-in data is consistent with the DMT specifiers.
# clean data additionally removes all voltages and only potentials remain
data = data.clean_data(["B", "C", "E"], reference_node="E")
# adjust unit from read in file (currents were in mA, DMT uses always basic units):
data[col_ic] = data[col_ic] * 1e-3
data[col_ib] = data[col_ib] * 1e-3
# The "ensure_specifier_column" method can be used to ensure that a given derived
# electrical quantity, such as a voltage, is inside the dataframe. If it is not there,
# an algorithm will try to generate the quantity.
data.ensure_specifier_column(col_vbe)
data.ensure_specifier_column(col_vbc)

# generate the plot object with proper axis and legend location
plot = Plot(
    "I_C(V_BE)",
    x_specifier=col_vbe,
    y_specifier=col_ic,
    y_scale=1e3,
    y_log=True,
    legend_location="lower right",
)
plot.legend_frame = False
plot.x_limits = (0.6, 1.1)

# fill the gummel plot with a line for each V_BC
for _index, vbc, data_vbc in data.iter_unique_col(col_vbc, decimals=3):
    plot.add_data_set(
        data_vbc[col_vbe],
        data_vbc[col_ic],
        label="${:s}= \\SI{{{:.1f}}}{{\\volt}}$".format(col_vbc.to_tex(), vbc),
    )

# show the plot (DMT has three plotting back-ends: matplotlib, pyqtgraph and tikz)
# plot.plot_pyqtgraph()
plot.plot_py()
# save the plot as tikz and build it
plot.save_tikz(
    path_static,
    standalone=True,
    build=True,
    clean=True,
    width="3in",
)



""" Provdes a class for TCAD DuTs

Provides a interface superclass. Here all methods which must be implemented by all TCAD interfaces are collected.

Author: Mario Krattenmacher | Mario.Krattenmacher@semimod.de
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import copy
import logging
from DMT.core import create_md5_hash, DutView


class DutTcad(DutView):
    """Superclass for common methods and attributes of TCAD DuTs

    Makes a TCAD simulator like DEVICE or Hdev useable by DMT

    Parameters
    ----------
    database_dir    : string
        This is the directory were the DUT will create its database.
    name      : string
        Prefix for the database
    dut_type   : :class:`~DMT.core.dut_type.DutType`
        Type of the DUT.
    nodes     : string
        Strings with comma separated node names of DUT. If nodes is None, nodes will be requested from Dut_type class.
    inp_structure : str or InpTcad
        One of 3 different possibilities: special obj for create_inp_header(), string with path to file or direct input file string
    simulator_command : str
        Command to start the correct TCAD simulator
    simulator_arguments : list[str]
        List of arguments for the simulator command, will be added one by one before the input file.

    Attributes
    ----------
    sim_command : str
        Command to start the correct TCAD simulator
    sim_args : list[str]
        List of arguments for the simulator command, will be added one by one before the input file.

    Methods
    -------
    inp_header(value)
        Returns input header of a simulation input file, or sets it to a new value, removing attached files.
    inp_structure(value)
        Returns input structure for DUT, or sets it while automatically creating a new header.
    get_hash()
        Returns hash if inp_header is set.
    get_start_sim_command()
        Returns command to start a specific type of simulation (DEVICE, Hdev, ...).
    run_sim(sweep)
        Starts a simulation.
    create_inp_header(inp_structure)
        Creates an inp_header for a given inp_structure. Has to be set by inheriting class.
    set_param()
        Enables setting model parameters for a simulation. Has to be set by in heriting class.
    get_param()
        Enables getting model parameters for a simulation. Has to be set by in heriting class.
    """

    def __init__(self, database_dir, name, dut_type, inp_structure, **kwargs):
        # use the property setter method to analyze inp_structure.
        self._inp_header = None
        self._inp_structure = None

        super().__init__(database_dir, name, dut_type, **kwargs)

        # set the input structure to the input header property and let it handle it.
        self.inp_header = inp_structure

    @property
    def inp_header(self):
        """Getter method for the input header of the simulation input file"""
        return self._inp_header

    @inp_header.setter
    def inp_header(self, value):
        """Setter method for the input header. This removes any stored data from the object, as it is not valid anymore!
        Automatically recalls create_inp_header, if it is not a string.

        Parameters
        ----------
        value : str or anything fitting for overwritten create_inp_header
        """
        self._data = {}
        self._inp_structure = None

        if isinstance(value, str):
            try:
                with open(value, "r") as inp_file:
                    self._inp_header = inp_file.read()
            except IOError:
                # could not open input file. Assume it is the already read string
                self._inp_header = value
        else:
            self._inp_structure = copy.deepcopy(value)
            self._inp_header = self.create_inp_header(value)

        logging.info(
            "Successfully created input header of dut %s%s!", self.name, str(self.get_hash())
        )
        logging.debug("Content:\n%s", self._inp_header)

    def get_hash(self):
        """Returns a md5 hash generated from self.inp_header, if it is not set, this will return False!

        In case a InpStructure is set, this always recreates the inp_header :/.

        Returns
        -------
        str or False
        """
        if self.inp_header is None:
            return False

        # i have to ensure always correct hash for a dut... sorry
        if self._inp_structure is not None:
            self._inp_header = self.create_inp_header(self._inp_structure)

        return create_md5_hash(self._inp_header + self.sim_command)

    def create_inp_header(self, inp_structure):
        """Creates the inp_header from the given parameter object.

        Parameters
        ----------
        inp_structure
            Type and content depends on implementation of inheriting class!

        Returns
        -------
        str
            Input circuit
        """
        raise NotImplementedError("create_inp_header() must be implemented in inheriting class!")



""" data processor module

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import numpy as np
from skrf import network as rf_network
from DMT.core import specifiers, set_col_name, sub_specifiers


def is_iterable(arg):
    """Returns True if the object is iterable

    Source: https://stackoverflow.com/a/36407550/13212532

    """
    try:
        _test = (e for e in arg)
        return True
    except TypeError:
        return False


def flatten(items):
    """Yield items from any nested iterable; see Reference https://stackoverflow.com/a/40857703."""
    for x in items:
        if not isinstance(x, (str, bytes)) and is_iterable(x):
            for sub_x in flatten(x):
                yield sub_x
        else:
            yield x


def strictly_increasing(L):
    """checks if given iterable is strictly increasing or not"""
    return all(x < y for x, y in zip(L, L[1:]))


class DataProcessor(object):
    """Basic class responsible for the manipulation and performing calculations on electrical Data in DMT.

    This class is designed as a `mixin <http://www.qtrac.eu/pyagg.html>`_ class. This design pattern allows to use multiple inheritance in order
    to extend the functionality of other DMT modules by the basic functions provided here.
    The functions here are basic electrical functions on np.array() objects with a clear syntax that should not change a lot in the future.

    Methods
    -------
    convert_n_port_para(para_values, p_from, p_to, z0=float(50)):
        Calculate the p_to n_port parameters from the p_from n_port parameters stored in para_values.

    deembed_short(s_para_values, s_para_short_values):
        de-embed the S parameters in s_para_values using the S parameters from the short dummy.

    deembed_open(s_para_values, s_para_open_values):
        de-embed the S parameters in s_para_values using the S parameters from the open dummy.

    calc_ft(freq, para_values, p_type):
        Calculate the transit frequency ft from the small signal parameters of type p_type in para_values.

    calc_fmax(freq, para_values, p_type):
        Calculate the maximum frequency of oscillation fmax from the small signal parameters of type p_type in para_values.

    calc_cbe(freq, para_values, p_type):
        Calculate the base-emitter junction capacitance from the small signal parameters of type p_type in para_values.

    calc_cbc(freq, para_values, p_type):
        Calculate the base-collector junction capacitance from the small signal parameters of type p_type in para_values.

    Notes
    -----
    ..todo: Nice tex equations here!
    """

    def fix_z0_shape(self, z0, nfreqs, nports):
        """
        Make a port impedance of correct shape for a given network's matrix

        This attempts to broadcast z0 to satisfy
            npy.shape(z0) == (nfreqs,nports)

        Parameters
        ----------
        z0 : number, array-like
            z0 can be:
            * a number (same at all ports and frequencies)
            * an array-like of length == number ports.
            * an array-like of length == number frequency points.
            * the correct shape ==(nfreqs,nports)

        nfreqs : int
            number of frequency points
        nports : int
            number of ports

        Returns
        -------
        z0 : array of shape ==(nfreqs,nports)
            z0  with the right shape for a nport Network

        Examples
        --------
        For a two-port network with 201 frequency points, possible uses may
        be

        >>> z0 = rf.fix_z0_shape(50 , 201,2)
        >>> z0 = rf.fix_z0_shape([50,25] , 201,2)
        >>> z0 = rf.fix_z0_shape(range(201) , 201,2)


        """

        if np.shape(z0) == (nfreqs, nports):
            # z0 is of correct shape. super duper.return it quick.
            return z0.copy()

        elif np.isscalar(z0):
            # z0 is a single number
            return np.array(nfreqs * [nports * [z0]])

        elif len(z0) == nports:
            # assume z0 is a list of impedances for each port,
            # but constant with frequency
            return np.array(nfreqs * [z0])

        elif len(z0) == nfreqs:
            # assume z0 is a list of impedances for each frequency,
            # but constant with respect to ports
            return np.array(nports * [z0]).T

        else:
            raise IndexError("z0 is not an acceptable shape")

    def s2y(self, s_in, z0=50):
        """
        convert scattering parameters [#]_ to admittance parameters [#]_


        .. math::
            y = \\sqrt {y_0} \\cdot (I - s)(I + s)^{-1} \\cdot \\sqrt{y_0}

        Parameters
        ----------
        s : complex array-like
            scattering parameters
        z0 : complex array-like or number
            port impedances

        Returns
        -------
        y : complex array-like
            admittance parameters

        See Also
        --------
        s2z
        s2y
        s2t
        z2s
        z2y
        z2t
        y2s
        y2z
        y2z
        t2s
        t2z
        t2y
        Network.s
        Network.y
        Network.z
        Network.t

        References
        ----------
        .. [#] http://en.wikipedia.org/wiki/S-parameters
        .. [#] http://en.wikipedia.org/wiki/Admittance_parameters
        """
        # cdef int nfreqs, nports
        s = np.copy(s_in)
        Id = np.zeros_like(s_in, dtype=np.cdouble)  # (nfreqs, nports, nports)
        sqrty0 = np.zeros_like(s_in, dtype=np.cdouble)  # (nfreqs, nports, nports)

        nfreqs = s_in.shape[0]
        nports = s_in.shape[1]

        z0 = self.fix_z0_shape(z0, nfreqs, nports)

        s[s == -1.0] = -1.0 + 1e-12  # solve numerical singularity
        s[s == 1.0] = 1.0 + 1e-12  # solve numerical singularity

        # The following is a vectorized version of a for loop for all frequencies.
        # Creating Identity matrices of shape (nports,nports) for each nfreqs
        np.einsum("ijj->ij", Id)[...] = 1.0
        # Creating diagonal matrices of shape (nports, nports) for each nfreqs
        np.einsum("ijj->ij", sqrty0)[...] = np.sqrt(1.0 / z0)
        # s -> y
        # y = sqrty0 @ (Id - s) @  npy.linalg.inv(Id + s) @ sqrty0  # Python>3.5
        y = np.matmul(np.matmul(np.matmul(sqrty0, (Id - s)), np.linalg.inv(Id + s)), sqrty0)
        return y

    def s2z(self, s_in, z0=50):
        """
        Convert scattering parameters [1]_ to impedance parameters [2]_


        .. math::
            z = \\sqrt {z_0} \\cdot (I + s) (I - s)^{-1} \\cdot \\sqrt{z_0}

        Parameters
        ----------
        s : complex array-like
            scattering parameters
        z0 : complex array-like or number
            port impedances.

        Returns
        -------
        z : complex array-like
            impedance parameters



        References
        ----------
        .. [1] http://en.wikipedia.org/wiki/S-parameters
        .. [2] http://en.wikipedia.org/wiki/impedance_parameters

        """
        nfreqs = s_in.shape[0]
        nports = s_in.shape[1]
        s = np.copy(s_in)
        Id = np.zeros_like(s_in, dtype=np.cdouble)  # (nfreqs, nports, nports)
        sqrtz0 = np.zeros_like(s_in, dtype=np.cdouble)  # (nfreqs, nports, nports)

        z0 = self.fix_z0_shape(z0, nfreqs, nports)

        s[s == -1.0] = -1.0 + 1e-12  # solve numerical singularity
        s[s == 1.0] = 1.0 + 1e-12  # solve numerical singularity

        # The following is a vectorized version of a for loop for all frequencies.
        # Creating Identity matrices of shape (nports,nports) for each nfreqs
        np.einsum("ijj->ij", Id)[...] = 1.0
        # Creating diagonal matrices of shape (nports, nports) for each nfreqs
        np.einsum("ijj->ij", sqrtz0)[...] = np.sqrt(z0)
        # s -> z
        # z = sqrtz0 @ npy.linalg.inv(Id - s) @ (Id + s) @ sqrtz0  # Python>3.5
        z = np.matmul(np.matmul(np.matmul(sqrtz0, np.linalg.inv(Id - s)), (Id + s)), sqrtz0)
        return z

    def convert_n_port_para(self, para_values, p_from, p_to, z0=float(50)):
        """n_port parameter conversion routine.

        Convert n_port parameters from p_from to p_to using scikit-rf. Available are all conversion between parameters S,Y,Z,T,A.

        Parameters
        ----------
        para_values  :  np.ndarray()
            Numpy array with shape (n_freq, n_port, n_port) holding the values of the small signal parameter p_from.

        p_from  :  string
            String that specifies which parameters are stored in p_from.

        p_to    :  string
            String that specifies which parameters should be created from the parameters p_from.

        z0      : float()
            Reference impedance in ohms, default is float(50).

        Returns
        -------
        para_new  :  np.ndarray()
            Numpy array with shape (n_freq, n_port, n_port) holding the values of the small signal parameter p_to.
        """
        if p_from == p_to:
            return para_values

        # call the right conversion routine in scikit-rf.
        if p_from == "S":
            if p_to == "Z":
                para_new = self.s2z(para_values, z0)
            elif p_to == "Y":
                para_new = self.s2y(para_values, z0)
            elif p_to == "T":
                para_new = rf_network.s2t(para_values)
            elif p_to == "A":
                para_new = rf_network.s2a(para_values, z0)
            elif p_to == "H":
                para_new = rf_network.s2y(para_values, z0)
                para_new = self.y2h(para_new)

        elif p_from == "Z":
            if p_to == "S":
                para_new = rf_network.z2s(para_values, z0)
            elif p_to == "Y":
                para_new = rf_network.z2y(para_values)
            elif p_to == "T":
                para_new = rf_network.z2t(para_values)
            elif p_to == "A":
                para_new = rf_network.z2a(para_values)

        elif p_from == "Y":
            if p_to == "S":
                para_new = rf_network.y2s(para_values, z0)
            elif p_to == "Z":
                para_new = rf_network.y2z(para_values)
            elif p_to == "T":
                para_new = rf_network.y2t(para_values)
            elif p_to == "A":  # i want to cry: y2a is not available in scikit-rf. workaround here
                para_new = self.y2a(para_values)

        elif p_from == "T":
            if p_to == "S":
                para_new = rf_network.t2s(para_values)
            elif p_to == "Z":
                para_new = rf_network.t2z(para_values)
            elif p_to == "Y":
                para_new = rf_network.t2y(para_values)

        elif p_from == "A":
            if p_to == "Y":
                para_new = self.a2y(para_values)
            elif p_to == "S":
                para_new = rf_network.a2s(para_values, z0)
            elif p_to == "Z":
                raise NotImplementedError

        else:
            raise IOError(
                "DMT -> DataProcessor: You specified an n-port Conversion that is not implemented or makes no sense."
            )

        return para_new

    def y2a(self, y):
        """My own y2a routine since scikit rf does not have one. What a shame!"""
        a = np.zeros(y.shape, dtype="complex")
        for fidx in range(y.shape[0]):
            det_y = np.linalg.det(y[fidx, :, :])

            a[fidx, 0, 0] = -y[fidx, 1, 1] / y[fidx, 1, 0]
            a[fidx, 0, 1] = -1.0 / y[fidx, 1, 0]
            a[fidx, 1, 0] = -det_y / y[fidx, 1, 0]
            a[fidx, 1, 1] = -y[fidx, 0, 0] / y[fidx, 1, 0]

        return a

    def y2h(self, y):
        """My own y2h routine since scikit rf does not have one. What a shame!"""
        h = np.zeros(y.shape, dtype="complex")
        for fidx in range(y.shape[0]):
            det_y = np.linalg.det(y[fidx, :, :])

            h[fidx, 0, 0] = 1
            h[fidx, 1, 1] = det_y
            h[fidx, 0, 1] = -y[fidx, 0, 1]
            h[fidx, 1, 0] = y[fidx, 1, 0]
            h[fidx, :, :] = h[fidx, :, :] / y[fidx, 0, 0]

        return h

    def a2y(self, a):
        """My own a2y routine since scikit rf does not have one. What a shame!"""
        y = np.zeros(a.shape, dtype="complex")
        for fidx in range(a.shape[0]):
            det_a = np.linalg.det(a[fidx, :, :])
            y[fidx, 0, 0] = 1.0 / a[fidx, 0, 1] * a[fidx, 1, 1]
            y[fidx, 1, 1] = 1.0 / a[fidx, 0, 1] * a[fidx, 0, 0]
            y[fidx, 0, 1] = -1.0 / a[fidx, 0, 1] * det_a
            y[fidx, 1, 0] = -1.0 / a[fidx, 0, 1]

        return y

    def parallel_norm(self, s_para_values, ndevices):
        """Normalize the measured S parameters in s_para_values to the number of parallel devices.

        Parameters
        ----------
        s_para_values       :  np.ndarray(np.cmplx128)
            S parameters that shall be de-embedded.

        ndevices            :  int
            Number of parallel devices.

        Returns
        -------
        s_para_values       :  np.ndarray(np.cmplx128)
            Normalized S-para-values

        """
        # create Y para from S para
        y_para_values = self.convert_n_port_para(
            s_para_values, "S", "Y"
        )  # (n_freq, n_port, n_port)
        # normalize Y para
        y_para_values = y_para_values / ndevices

        return self.convert_n_port_para(y_para_values, "Y", "S")

    def calc_RBC_RBE(self, mres, df_RM):
        """Calculate the metallization resistances R_CE and R_BE.

        Parameters
        ----------
        df_RM   : df
            Containing all values for the incoming df, where V_CE=0V.

        Returns
        -------
        mres    : [list]
            List containing the calculated resistance values
        """

        df_RM_clean = df_RM.dropna(axis=0, how="any")

        R_BCM = np.polyfit(np.negative(df_RM_clean["I_C"]), df_RM_clean["V_B"], 1)

        ic_2 = df_RM_clean["I_B"] + df_RM_clean["I_C"]
        R_BEM = np.polyfit(ic_2, df_RM_clean["V_B"], 1)

        mres["R_BCM"] = R_BCM[0]
        mres["R_BEM"] = R_BEM[0]
        return mres

    def calc_RCE(self, mres, df_RM):
        """Calculate the metallization resistances R_BC.

        Parameters
        ----------
        df_RM   : df
            Containing all values for the incoming df, where V_BE=0V.

        Returns
        -------
        mres    : {dict}
            List containing the calculated resistance values
        """
        df_RM_clean = df_RM.dropna(axis=0, how="any")

        R_CEM = np.polyfit((df_RM_clean["I_B"] + df_RM_clean["I_C"]), df_RM_clean["V_C"], 1)

        mres["R_CEM"] = R_CEM[0]

        return mres

    def convert_mres(self, mres):
        """Converts the calculated resistance network from delta- to wye-form.

        Parameters
        ----------
        mres    : {dict}
            List containing the calculated delta-form resistances
        Returns
        -------
        mres    : {dict}
            Same list with appended wye-form parameters.
        """
        R_sum = mres["R_BCM"] + mres["R_CEM"] + mres["R_BEM"]

        R_BM = (mres["R_BEM"] * mres["R_BCM"]) / R_sum
        R_CM = (mres["R_BCM"] * mres["R_CEM"]) / R_sum
        R_EM = (mres["R_BEM"] * mres["R_CEM"]) / R_sum

        mres["R_BM"] = R_BM
        mres["R_CM"] = R_CM
        mres["R_EM"] = R_EM

        return mres

    def deembed_mres(self, df, mres):
        """Substract external voltage drop over metal resistances from measured voltages.

        Parameters
        ----------
        df      : DMT.dataframe
            df contains DC measurements that are to be deembedded.
        mres    : {'R_BM':float64, 'R_CM':float64, 'R_EM'_float64}
            List of calculated resistances R_CM, R_BM, and R_EM.
        """
        col_vb, col_ib, col_vc, col_ic, col_ve = None, None, None, None, None
        try:
            col_vb = df.get_col_name(specifiers.VOLTAGE, "B")
            col_vb_force = set_col_name(
                specifiers.VOLTAGE, "B", sub_specifiers=sub_specifiers.FORCED
            )
            df[col_vb_force] = df[col_vb].to_numpy()
        except KeyError:
            pass
        try:
            col_vc = df.get_col_name(specifiers.VOLTAGE, "C")
            col_vc_force = set_col_name(
                specifiers.VOLTAGE, "C", sub_specifiers=sub_specifiers.FORCED
            )
            df[col_vc_force] = df[col_vc].to_numpy()
        except KeyError:
            pass
        try:
            col_ve = df.get_col_name(specifiers.VOLTAGE, "E")
            col_ve_force = set_col_name(
                specifiers.VOLTAGE, "E", sub_specifiers=sub_specifiers.FORCED
            )
            df[col_ve_force] = df[col_ve].to_numpy()
        except KeyError:
            pass
        try:
            col_ic = df.get_col_name(specifiers.CURRENT, "C")
        except KeyError:
            pass
        try:
            col_ib = df.get_col_name(specifiers.CURRENT, "B")
        except KeyError:
            pass

        if col_vb and col_ib:
            vb = df[col_vb].to_numpy()
            ib = df[col_ib].to_numpy()
            # vb_dif      = [i * mres["R_BM"] for i in ib]
            vb_dif = ib * mres["R_BM"]
            df[col_vb] = vb - vb_dif

        if col_vc and col_ic:
            vc = df[col_vc].to_numpy()
            ic = df[col_ic].to_numpy()
            vc_dif = ic * mres["R_CM"]
            df[col_vc] = vc - vc_dif

        if col_ve and col_ib and col_ic:
            ve = df[col_ve].to_numpy()
            # ie = [sum(i) for i in zip(ic, ib)]
            ie = ib + ic
            # ve_dif      = [i * mres["R_EM"] for i in ie]
            ve_dif = ie * mres["R_EM"]
            df[col_ve] = ve + ve_dif

        return df

    def deembed_short(self, s_para_values, s_para_short_values, times=1):
        """Deembed the measured S parameters in s_para_values from the measured S parameters in s_para_short_values.

        Parameters
        ----------
        s_para_values       :  np.ndarray(np.cmplx128)
            S parameters that shall be de-embedded.

        s_para_short_values  :  np.ndarray(np.cmplx128)
            S parameters of a short dummy.

        times               : int
            Number of times the short has to be removed

        Returns
        -------
        s_para_values       :  np.ndarray(np.cmplx128)
            Short de-embedded S parameters.
        """
        # create Z para from S para
        z_para_values = self.convert_n_port_para(s_para_values, "S", "Z")
        z_para_short_values = self.convert_n_port_para(s_para_short_values, "S", "Z") / times

        # find number of frequency sweeps
        n_sweeps = z_para_values.shape[0] / z_para_short_values.shape[0]
        if not np.mod(n_sweeps, 1) == 0:
            raise IOError(
                "DMT -> DataProcessor: short de-embedding structure does not match the data in df."
            )

        z_para_short_values = np.tile(z_para_short_values, [int(n_sweeps), 1, 1])

        # de-embed
        z_para_values = z_para_values - z_para_short_values

        # convert back to S para
        return self.convert_n_port_para(z_para_values, "Z", "S")

    def deembed_open(self, s_para_values, s_para_open_values, times=1):
        """Deembed the measured S parameters in s_para_values from the measured S parameters in s_para_open_values.

        Parameters
        ----------
        s_para_values       :  np.ndarray(np.cmplx128)
            S parameters that shall be de-embedded.

        s_para_open_values  :  np.ndarray(np.cmplx128)
            S parameters of a short dummy.

        times               : int
            Numer of times the open has to be removed.

        Returns
        -------
        s_para_values       :  np.ndarray(np.cmplx128)
            Short de-embedded S parameters.
        """
        # calculate the y parameters in both dataframes
        y_para_values = self.convert_n_port_para(s_para_values, "S", "Y")
        y_para_open_values = self.convert_n_port_para(s_para_open_values, "S", "Y")

        # find the number of frequency sweeps
        n_sweeps = y_para_values.shape[0] / y_para_open_values.shape[0]
        if not np.mod(n_sweeps, 1) == 0:
            raise IOError(
                "DMT -> DataProcessor: open de-embedding structure does not match the data in df."
            )

        y_para_open_values = np.tile(y_para_open_values, [int(n_sweeps), 1, 1])
        # deembed
        y_para_values = y_para_values - y_para_open_values * times

        # convert Y para back to S para
        return self.convert_n_port_para(y_para_values, "Y", "S")

    def calc_ft(self, freq, para_values, p_type):
        """Calculate the transit frequency F_T using the spot frequency method.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        F_T  :  np.ndarray()
            Array of shape [n_freq] that contains ft from the spot frequency method.
        """
        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")

        # calculate ft = f / im( Y(1,1)/Y(2,1) ) element-wise for every frequency
        return freq / np.imag(y_para_values[:, 0, 0] / y_para_values[:, 1, 0])

    def calc_tfit1(self, freq, para_values, p_type):
        """Calculate the transit frequency F_T using the spot frequency method.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        tfit  :  np.ndarray()
            Array of shape [n_freq] that contains tfit from the spot frequency method.
        """

        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")
        omega = 2 * np.pi * freq
        gm = np.real(y_para_values[:, 1, 0])
        # tfi = im ( Y_11 + Y_21)/(gm*omega)
        return np.imag(y_para_values[:, 0, 0] + y_para_values[:, 0, 1]) / (omega * gm)

    def calc_tfit2(self, freq, para_values, p_type):
        """Calculate the transit frequency F_T using the spot frequency method.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        tfit  :  np.ndarray()
            Array of shape [n_freq] that contains tfit from the spot frequency method.
        """

        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")
        omega = 2 * np.pi * freq
        gm = np.real(y_para_values[:, 1, 0])
        # tfi = im ( Y_11 + Y_21)/(gm*omega)
        return np.imag(y_para_values[:, 0, 0] + y_para_values[:, 1, 0]) / (omega * gm)

    def calc_fmax(self, freq, para_values, p_type):
        """Calculate the maximum frequency of oscillation from the unilateral gain.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        fmax         :  np.ndarray()
            Array of shape [n_freq] that contains ft from the spot frequency method.
        """
        # calc the unilateral gain
        gu = self.calc_unilateral_gain(freq, para_values, p_type)

        # calculate FMAX
        return freq * gu**0.5

    # pylint: disable=unused-argument
    def calc_unilateral_gain(self, freq, para_values, p_type):
        """Calculates the unilateral gain `GU <https://www2.eecs.berkeley.edu/Pubs/TechRpts/2016/EECS-2016-15.pdf>`_ .
        | https://ieeexplore.ieee.org/stamp/stamp.jsp?arnumber=1083579

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        gu           :  np.ndarray()
            Unilateral gain as an array with shape [n_freq].
        """
        # from y parameters
        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")
        # calculate GU
        gu = (
            abs(y_para_values[:, 1, 0] - y_para_values[:, 0, 1]) ** 2
            / 4.0
            / (
                np.real(y_para_values[:, 0, 0]) * np.real(y_para_values[:, 1, 1])
                - np.real(y_para_values[:, 0, 1]) * np.real(y_para_values[:, 1, 0])
            )
        )
        return gu

    def calc_mag(self, freq, para_values, p_type):
        """Calculates the maximum available gain MAG.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        mag           :  np.ndarray()
            maximum available gain mag.
        """
        s_para_values = self.convert_n_port_para(para_values, p_type, "S")

        k = self.calc_k(freq, s_para_values, "S")
        mag = (
            (k - np.sqrt(k**2.0 - 1.0))
            * np.abs(s_para_values[:, 1, 0])
            / np.abs(s_para_values[:, 0, 1])
        )

        return mag

    def calc_k(self, freq, para_values, p_type):
        """Calculates the k-factor.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        k           :  np.ndarray()
            Stability factor k.

        Notes
        -----
        ..todo: from s para
        """
        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")
        k = (
            2 * np.real(y_para_values[:, 0, 0]) * np.real(y_para_values[:, 1, 1])
            - np.real(y_para_values[:, 0, 1] * y_para_values[:, 1, 0])
        ) / np.abs(y_para_values[:, 0, 1] * y_para_values[:, 1, 0])
        return k

    def calc_msg(self, freq, para_values, p_type):
        """Calculates the maximum stable gain MSG.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        msg          :  np.ndarray()
            Maximum stable gain MAG.

        Notes
        -----
        ..todo: calc from S para too
        """
        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")

        msg = np.absolute(y_para_values[:, 1, 0]) / np.absolute(y_para_values[:, 0, 1])
        return msg

    def calc_cap_shunt_port_1(self, freq, para_values, p_type):
        """Calculates the shunt capacitance at port 1 assuming a Pi equivalent circuit capacitance cbe from the small signal parameters para_values.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        cbe           :  np.ndarray()
            Base emitter junction capacitance as an array with shape [n_freq].
        """
        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")

        # calculate cbe from y para
        cbe = np.imag(y_para_values[:, 0, 0] + y_para_values[:, 0, 1]) / (2.0 * np.pi * freq)

        return cbe

    def calc_cap_series_thru(self, freq, para_values, p_type):
        """Calculates the the series-thru junction capacitance cbc from the small signal parameters para_values.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        cbc           :  np.ndarray()
            Base collector junction capacitance as an array with shape [n_freq].
        """
        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")

        # calculate GU
        return -np.imag(y_para_values[:, 0, 1]) / (2.0 * np.pi * freq)

    def calc_cap_shunt_port_2(self, freq, para_values, p_type):
        """Calculates the active shunt capacitance at port_2 assuming a PI equivalent circuit.

        Parameters
        ----------
        freq         :  np.ndarray()
            Frequencys that correspond to para_values.
        para_values  :  np.ndarray()
            Small signal parameters of type p_type with shape [n_freq, n_port, n_port]
        p_type       :  string
            Type of the small signal parameters in para_values.

        Returns
        -------
        cce           :  np.ndarray()
            Collector emitter junction capacitance as an array with shape [n_freq].
        """
        y_para_values = self.convert_n_port_para(para_values, p_type, "Y")

        # calculate cce from y para
        cce = np.imag(y_para_values[:, 1, 1] + y_para_values[:, 0, 1]) / (2.0 * np.pi * freq)

        return cce

    def calc_beta(self, ic, ib):
        """Calculates the base collector current amplification from dc currents

        Parameters
        ----------
        ic  :  np.ndarray()
            DC Collector current.
        ib  :  np.ndarray()
            DC Base current.

        Returns
        -------
        beta  :  np.ndarray()
            Base collector junction capacitance as an array with shape [n_freq].
        """
        return ic / ib  # I know this seems crazy, however it ist clean coding style.

    def calc_gm(self, ic, vbe, vb_forced=None, vc_forced=None):
        """Calculates the transconductance of a bjt or mosfet.

        Parameters
        ----------
        ic  :  np.ndarray()
            DC Collector current.
        vbe  :  np.ndarray()
            be voltage
        vbc  :  np.ndarray()
            bc voltage

        Returns
        -------
        gm  :  np.ndarray()
            The transconductance
        """
        # vb_forced is not necessarily increasing as needed for gm calculation, e.g. vb is not increasing at constant vc.
        # following code tries to catch that.

        gm = np.zeros_like(vbe)
        if vb_forced is not None and vc_forced is not None:
            vbc_forced = vb_forced - vc_forced
            # V_BC or V_C forced?
            vc_forced_unique = np.unique(np.round(vc_forced, decimals=3))
            vbc_forced_unique = np.unique(np.round(vbc_forced, decimals=3))
            vbc_unique = vbc_forced_unique.size < vc_forced_unique.size
            if vbc_unique:
                v_outer_unique = vbc_forced_unique
                v_outer = vbc_forced
            else:
                v_outer_unique = vc_forced_unique
                v_outer = vc_forced

            for v_out in v_outer_unique:
                indices = np.isclose(v_outer, v_out, rtol=1e-3, atol=4.99e-3)
                vb_unique, indicies_unique, unique_inverse = np.unique(
                    vb_forced[indices], return_index=True, return_inverse=True
                )
                sorted_indices = np.argsort(vb_unique)
                ic_unique = ic[indices][indicies_unique][sorted_indices]
                res = np.empty_like(ic_unique)
                res[sorted_indices] = (
                    np.gradient(np.log(np.abs(ic_unique)), vb_unique[sorted_indices], edge_order=2)
                    * ic_unique
                )
                gm[indices] = res[unique_inverse]

        elif vc_forced is None and vb_forced is None:
            resort = None
            # sort to increasing vbe
            if not strictly_increasing(vbe):
                vbe_indexes = vbe.argsort()
                resort = vbe_indexes.argsort()
                vbe = vbe[vbe_indexes]
                ic = ic[vbe_indexes]

            # calculate
            gm = np.gradient(np.log(np.abs(ic)), vbe) * ic

            if resort is not None:
                gm = gm[resort]

        else:
            raise NotImplementedError(
                "DMT data_processor -> combination of arguments not yet implemented."
            )

        return np.where(gm <= 0.0, np.nan, gm)



""" Testing McParameter, McParameterCollection, MCard and McHicum from L2
"""
from importlib.resources import read_text
import os
import logging
import pytest
import numpy as np
from pathlib import Path
from DMT.core import MCard

from DMT.core.mc_parameter import McParameter
from DMT.core.circuit import HICUML2_HBT

from DMT.exceptions import BoundsError, ValueTooLargeError

folder_path = Path(__file__).resolve().parent
hicum_path = folder_path / "test_va_code" / "hicuml2" / "hicumL2V2p4p0_release.va"
logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_mcard_class.log",
    filemode="w",
)


def test_type_mc_parameter():
    """Test the type (float or int) of the mc_parameter class"""
    a = McParameter("asd", value=int(65))  # setting a integer value, but not, float type
    assert (
        a.val_type == float
    )  # even if a integer is set, the type is still float to prevent unwanted conversions
    assert a.value == int(65)

    a.value = 65.43

    with pytest.raises(IOError):
        a.val_type = int

    a.value = 65  # here we set a integer value and int type
    a.val_type = int

    with pytest.raises(TypeError):  # can not set a float number anymore
        a.value = 65.342

    with pytest.raises(BoundsError):  # can not set a maximum below the current value
        a.max = 20

    a.max = 100
    with pytest.raises(ValueTooLargeError):  # can not set a value above the maximum value
        a.value = 120


def test_format_mc_parameter():
    """Format mc parameters, so this is just pretty printing feature!

    Notice: __format__ is redefined in class McParameter in which format "s" corresponds name in McParameter, and "d","e","f", "g" corredpond value
    """

    value = 6556.1234
    name = "para"
    a = McParameter(name, value=value)
    assert "{0:f}".format(a) == "{0:f}".format(value)
    assert "{0:e}".format(a) == "{0:e}".format(value)

    assert "{0:s}".format(a) == "{0:s}".format(name)


def test_mcard_load_compare():
    """Compare two loaded model cards"""
    mc_D21 = MCard(
        ["C", "B", "E", "S", "T"],
        default_module_name="",
        default_subckt_name="",
    )
    mc_D21.load_model_parameters(
        folder_path / "test_modelcards" / "IHP_ECE704_03_para_D21.mat",
    )
    mc_N5 = MCard(
        ["C", "B", "E", "S", "T"],
        default_module_name="",
        default_subckt_name="",
    )
    mc_N5.load_model_parameters(
        folder_path / "test_modelcards" / "ITRS_N5.mat",
    )
    assert mc_N5 != mc_D21


def test_set_va_code():

    mc = MCard(
        ["A"],
        "Q_HIC",
        "hicuml2va",
        va_file=hicum_path,
    )

    assert mc.va_codes is not None and (
        all([vafile in [hicum_path.name] for (vafile, vacode) in mc.va_codes.iter_codes()])
        and len(mc.va_codes) == 1
    )
    assert set(mc.nodes_list) == {"c", "b", "e", "tnode", "s"}

    mc.set_va_codes(
        folder_path / "test_va_code" / "diode_cmc_160823" / "diode_cmc.va",
    )

    va_codes_target_names = {
        "diode_cmc.va",
        "DIODE_CMC_InitModel.include",
        "DIODE_CMC_macrodefs.include",
        "DIODE_CMC_parlist.include",
        "DIODE_CMC_SIMKIT_macrodefs.include",
        "DIODE_CMC_varlist1.include",
        "DIODE_CMC_varlist2.include",
    }

    assert set(vafile for vafile, vacode in mc.va_codes.iter_codes()) == va_codes_target_names

    file_name = "test_va_codes.json"
    mc.dump_json(file_name, save_va_code=False)

    mc_saved = MCard.load_json(file_name)
    assert mc_saved.va_codes is None

    mc.dump_json(file_name)  # is true per default
    mc_saved = MCard.load_json(file_name)
    assert mc_saved.va_codes is not None and (
        set(vafile for vafile, vacode in mc_saved.va_codes.iter_codes()) == va_codes_target_names
    )

    file_name_compressed = "test_va_codes_compressed.json"
    mc.dump_json(file_name_compressed, compress_va_code=True)
    mc_saved = MCard.load_json(file_name_compressed)
    assert mc_saved.va_codes is not None and (
        set(vafile for vafile, vacode in mc_saved.va_codes.iter_codes()) == va_codes_target_names
    )
    assert mc_saved.va_codes == mc.va_codes

    os.remove(file_name)
    os.remove(file_name_compressed)

    # test the regex for minimal working environment..
    mc_saved.read_va_file_boundaries()

    assert mc_saved.default_module_name == "DIODE_CMC"
    assert len(mc_saved.paras) == 102
    assert set(mc_saved.nodes_list) == {"A", "K"}

    # test the vae updating
    assert "c10" in mc.name
    mc.update_from_vae(remove_old_parameters=True)

    assert "c10" not in mc.name
    assert set(mc.nodes_list) == {"A", "K"}


def test_read_va_file():
    mc = MCard(["A"], "Q_HIC", "hicuml2va", va_file=hicum_path)
    mc.update_from_vae()  # McHicum(va_file=VA_FILES["L2V2.4.0_release"])

    assert len(mc) == 136


def test_json():
    mc = MCard(["A"], "Q_HIC", "hicuml2va", va_file=hicum_path)
    mc.set_values({"c10": 3e-15})  # just a little change

    file_name = "test_json_mchicum.json"
    mc.dump_json(file_name)
    mc_read = MCard.load_json(file_name)

    assert mc == mc_read

    mc_read_general = MCard.load_json(file_name)  # This should also work?
    assert mc == mc_read_general

    os.remove(file_name)

    mc = MCard(["A"], "Q_HIC", "hicuml2va", va_file=hicum_path)
    with pytest.raises(KeyError):
        mc.set_values({"is": 3e-15})  # just a little change


def test_load_MCard_v1():
    """Load a version 1 model"""
    file_name = folder_path / "test_modelcards" / "N3_MCard_v1.json"
    mc_read = MCard.load_json(file_name)

    assert mc_read.va_codes is None  # path can not be resolved!

    assert np.isclose(mc_read["c10"].value, 5.991798808405992e-30)


if __name__ == "__main__":
    test_type_mc_parameter()
    test_format_mc_parameter()
    test_mcard_load_compare()
    test_set_va_code()
    test_read_va_file()
    test_json()
    test_load_MCard_v1()



""" Manage measurement data in DMT in as a measurement dut. Additionally has attributes which define the size and die of the dut.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

import logging
import re
from DMT.core import print_progress_bar
from DMT.core.dut_view import DutView


class DutMeas(DutView):
    """Subclass of DutView that is used to capture relevant methods and properties of measured data.

    Parameters
    ----------
    database_dir         :  string
        Root directory for saving duts and databases of this project.
    name           :  string
        Name of the dut to save it human readable.
    dut_type       : :class:`~DMT.core.dut_type.DutType`
    width          :  float64, optional
        Width in m.
    length         :  float64, optional
        Length in m.
    wafer          :  string, optional
        Wafer identification string of this object.
    die            :  string, optional
        Die identification string of this object.
    ndevices       :  integer
        Number of physically parallel devices of this measurement.
    hash           :  float64
        Unique hash of this object.

    Attributes
    ----------
    width          :  float64
        Width in m.
    length         :  float64
        Length in m.
    perimeter      :  float64
        Perimeter in m.
    area           :  float64
        Area of this DutMeas object in m^2.
    wafer          :  string
        Wafer identification string of this object.
    die            :  string
        Die identification string of this object.
    ndevices       :  integer
        Number of physically parallel devices of this measurement.

    Methods
    -------
    run_simulation(sweep)
        Return a NotImplemented exception. Measurements can not be simulated!
    get_hash()
        Return a unique hash based on some attributes of the DutMeas object.
    add_short(short, key='/dummies/short', force=True)
        Add a short dummy to the database of this DutMeas object.
    add_open(short, key='/dummies/short', force=True)
        Add an open dummy to the database of this DutMeas object.
    deembed(filter_func, short_key='/dummies/short', open_key='/dummies/open')
        Deembed some dataframes of this DutMeas object using a user supplied filter function.
    clean_data()
        Clean the dataframe columns of the DataFrame objects in this DutMeas objects database.
    get_output_data(sweep)
        Raises NotImplementedError since a measurement can not return its output files.

    """

    def __init__(
        self,
        database_dir,
        name,
        dut_type,
        force=False,
        wafer=None,
        die=None,
        ndevices=1,
        deemb_name=None,
        **kwargs,
    ):
        super().__init__(database_dir, name, dut_type, nodes=None, force=force, **kwargs)
        self.wafer = wafer
        self.die = die
        self.ndevices = ndevices

        if deemb_name is None:
            self.deemb_name = self.name
        else:
            self.deemb_name = deemb_name

        if self.width is not None:
            try:
                self.perimeter = (self.width + self.length) * 2
                self.area = self.width * self.length
            except TypeError:
                pass  # if either spacer is None or width or length are tuples.

    def run_simulation(self, sweep):
        """Raise a OSError. Measurements can not be simulated!

        Raises
        ------
        OSError
            Always when called.
        """
        raise OSError("DutMeas: Measured data can not be simulated!")

    def prepare_simulation(self, sweep):
        """Raise a OSError. Measurement duts can not be simulated!

        Raises
        ------
        OSError
            Always when called.
        """
        raise OSError("DutMeas: Measured data can not be simulated!")

    def make_input(self, sweep):
        """Should genrate the input file for a given simulation sweep. This is not possible for measurements!

        Raises
        ------
        OSError
            Always when called.
        """
        raise OSError("DutMeas: Measured data can not be simulated!")

    def get_hash(self):
        """Empty string, so it is evaluated to false, but still a string.

        Returns
        -------
        ''
        """
        return ""

    def __format__(self, key):
        if key == "":
            return f"DutMeas:\n\tname: {self.name}\n\ttype: {self.dut_type}\n\tcontact_config: {self.contact_config}\n\twidth: {self.width}\n\tlength: {self.length}\n\tndev: {self.ndevices}"
        elif key == "long":
            return f"DutMeas:\n\tname: {self.name}\n\ttype: {self.dut_type}\n\tcontact_config: {self.contact_config}\n\twidth: {self.width}\n\tlength: {self.length}\n\tndev: {self.ndevices}\n\t data: {[key for key in self.data.keys]}"
        else:
            raise ValueError(f"Unkown format {key}")

    def add_short(self, short, key=r"dummies/short"):
        """Add a short dummy to the database of this DutMeas object.

        Parameters
        ----------
        short  :  string or DMT.DataFrame or pd.DataFrame
            The File or DataFrame with the short small signal parameters as a function of frequency.

        key    :  string
            Key that will be used to save the short dummy in the database. Default='/dummies/short'

        """

        self.add_data(short, key, force=True)
        logging.info("DMT -> DutMeas -> add_short(): Added a short to the DUT")

    def add_open(self, open_, key=r"dummies/open"):
        """Add an open dummy to the database of this DutMeas object.

        Parameters
        ----------
        open_  :  string or DMT.DataFrame or pd.DataFrame
            The File or DataFrame with the short small signal parameters as a function of frequency.

        key    :  string
            Key that will be used to save the short dummy in the database. Default='/dummies/short'

        """
        self.add_data(open_, key, force=True)
        logging.info("DMT -> DutMeas -> add_open(): Added an open to the DUT")

    def deembed(self, filter_func, short_key="dummies/short", open_key="dummies/open"):
        """Deembed some dataframes of this DutMeas object using a user supplied filter function.

        Parameters
        ----------
        filter_func  :  callable object
            This used supplied callable object, e.g. function, is called with each key in the DutView's database.
            If True is returned, the dataframe corresponding to key is deembedded.
        short_key    :  string
            Default='/dummies/short'. Key that corresponds to the short dummy in the database.
        open_key     :  string
            Default='/dummies/open'. Key that corresponds to the open dummy in the database.

        Returns
        -------
        self  :  DutMeas
            The updated DutMeas object.

        Notes
        -----
        todo: Delete this method since this functionality will be taken over by DutLib!
        """
        raise IOError("Obsolete")

        # retrieve the short and open dummies
        no_dummies = False
        try:
            df_short = self.data[short_key]
            df_open = self.data[open_key]
        except:
            no_dummies = True

        # find keys of dataframes that need to be deembedded
        keys = [key for key in self.data.keys() if filter_func(key)]

        if len(keys) > 0 and no_dummies:
            raise IOError(
                "DMT -> DutMeas -> deembed: Found keys that require deembedding, however no short or open dummies are present in this Dut."
            )

        if len(keys) == 0:  # some DUTs are only DC
            return self

        # go through each DataFrame where the function filter returns true and deembed()
        i = 0
        n_frames = len(self.data)
        print_progress_bar(i, n_frames, prefix="Deembedding:", suffix="Complete\n", length=50)
        for key in self.data.keys():
            if filter_func(key):
                i = i + 1
                df = self.data[key]
                self.data[key] = df.deembed(df_open, df_short, ports=self.nodes)
                logging.info("DMT -> DutMeas -> deembed(): Deembedded dataframe with key %s.", key)
                print_progress_bar(
                    i, n_frames, prefix="Deembedding:", suffix="Complete\n", length=50
                )

        if (
            self.ndevices > 1
        ):  # if we have more than one device in parallel, normalize the currents to the area of one device
            for key in self.data.keys():
                df = self.data[key]
                self.data[key] = df.parallel_norm(self.ndevices)

        return self

    def get_output_data(self, sweep):
        raise NotImplementedError(
            "DMT -> DutMeas -> get_output_data(): DutMeas does not implement this method on purpose!"
        )

    def join_key_temperature(self, *key_parts, temperature_converter=None):
        """Collects the temperature from key_parts and replaces it by a proper temperature key part.

        Parameters
        ----------
        key_parts : [str]
            List of key strings to join. One element contains the temperature information.
        temperature_converter : callable object, optional
            Function to convert a key part into a temperature.

        Returns
        -------
        key : str
            The correctly joined key_parts. If the temperature was inside, it is converted into "T" + temp + "K".

        """

        if temperature_converter is None:
            temperature_converter = self.temp_converter_default

        key_parts = list(key_parts)
        for i_key_part, key_part in enumerate(key_parts):
            temp = temperature_converter(key_part)
            if isinstance(temp, str):
                # escape to directly manipulate the key
                key_parts[i_key_part] = temp
                continue
            elif temp == -1:  # this value is returned it the temperature is not in the key.
                continue
            elif temp < 0:
                raise ValueError(
                    "Got a negative temperature for the key. This is nonphysical since the temperature must be in Kelvin!"
                )
            elif temp < 50:
                warning = (
                    "DutMeas -> join_key_temperature: The temperature of a measurement is below 50K. The given measurement is "
                    + self.join_key(*key_parts)
                    + " of the DuT "
                    + self.name
                    + "."
                )
                logging.warning(warning)
                print(warning)

            key_parts[i_key_part] = "T{:.2f}K".format(temp)

        return self.join_key(*key_parts)

    def temp_converter_default(self, key_part):
        """Default converter from key_part to temperature. Asumes the temperature in the key parts is in Kelvin.

        Parameters
        ----------
        key_part : str
            Key to inspect

        Returns
        -------
        temp : float or None
            If key does not contain a temperature, -1 is returned.
            If id does contain a temperature, the temperature in Kelvin is returned.
        """
        temp = None
        re_temp = re.search(r"T([0-9p\.]+)K", key_part)
        if re_temp:
            try:
                # always replace "p" with ".", if it is already with ".", it doesn't matter
                temp = round(float(re_temp.group(1).replace("p", ".")), 3)
            except ValueError:
                # if a value error in the except clause happens, try the next key part.
                pass

        # alternative:
        if key_part.startswith("T"):
            try:
                temp = round(float(key_part[1:].replace("p", ".")), 3)
            except ValueError:
                # if a value error in the except clause happens, try the next key part.
                pass

        # finally as a last escape: direct conversion :(
        try:
            temp = round(float(key_part.replace("p", ".")), 3)
        except ValueError:
            # if a value error in the except clause happens, try the next key part.
            pass

        if temp is None:
            return -1
        else:
            return temp



from numpy import array
import pytest
from DMT.core import DataFrame, specifiers

specifier_vbe = specifiers.VOLTAGE + "B" + "E"
specifier_vce = specifiers.VOLTAGE + "C" + "E"
specifier_vbc = specifiers.VOLTAGE + "B" + "C"
potential_b = specifiers.VOLTAGE + "B"
potential_c = specifiers.VOLTAGE + "C"
potential_e = specifiers.VOLTAGE + "E"


def test_2_voltages_0_potential_and_reference():
    df = DataFrame.from_dict({specifier_vbe: [0.5], specifier_vce: [1]})
    df.ensure_specifier_column(potential_b, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_c, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_e, debug=True, reference_node="E")
    vb = df[potential_b].to_numpy()
    vc = df[potential_c].to_numpy()
    ve = df[potential_e].to_numpy()
    assert vb == array([0.5])
    assert vc == array([1])
    assert ve == array([0])


def test_2_voltages_0_potential_and_reference_b_common1():
    df = DataFrame.from_dict({specifier_vbe: [0.5], specifier_vbc: [-0.5]})
    df.ensure_specifier_column(potential_b, debug=True, reference_node="B")
    df.ensure_specifier_column(potential_c, debug=True, reference_node="B")
    df.ensure_specifier_column(potential_e, debug=True, reference_node="B")
    vb = df[potential_b].to_numpy()
    vc = df[potential_c].to_numpy()
    ve = df[potential_e].to_numpy()
    assert vb == array([0])
    assert vc == array([0.5])
    assert ve == array([-0.5])


def test_2_voltages_0_potential_and_reference_b_common2():
    df = DataFrame.from_dict({specifier_vbe: [0.5], specifier_vbc: [-0.5]})
    df.ensure_specifier_column(potential_e, debug=True, reference_node="B")
    df.ensure_specifier_column(potential_c, debug=True, reference_node="B")
    df.ensure_specifier_column(potential_b, debug=True, reference_node="B")
    vb = df[potential_b].to_numpy()
    vc = df[potential_c].to_numpy()
    ve = df[potential_e].to_numpy()
    assert vb == array([0])
    assert vc == array([0.5])
    assert ve == array([-0.5])


def test_2_voltages_0_potential_and_reference_b_common3():
    df = DataFrame.from_dict({specifier_vbe: [0.5], specifier_vce: [-0.5]})
    df.ensure_specifier_column(potential_e, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_c, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_b, debug=True, reference_node="E")
    vb = df[potential_b].to_numpy()
    vc = df[potential_c].to_numpy()
    ve = df[potential_e].to_numpy()
    assert vb == array([0.5])
    assert vc == array([-0.5])
    assert ve == array([0])


def test_2_voltage_1_potential_and_reference():
    ## this should work
    df = DataFrame.from_dict({specifier_vbe: [-0.5], specifier_vbc: [0.5], potential_c: [0]})
    df.ensure_specifier_column(potential_b, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_c, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_e, debug=True, reference_node="E")
    assert df[potential_e].to_numpy() == array([0])
    assert df[potential_b].to_numpy() == array([-0.5])
    assert df[potential_c].to_numpy() == array([-1.0])


def test_1_voltage_1_potential_and_reference():
    ## this should work
    df = DataFrame.from_dict({specifier_vbe: [0.5], potential_c: [1]})
    df.ensure_specifier_column(potential_c, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_e, debug=True, reference_node="E")
    df.ensure_specifier_column(potential_b, debug=True, reference_node="E")
    assert df[potential_b].to_numpy() == array([0.5])
    assert df[potential_e].to_numpy() == array([0])
    assert df[potential_c].to_numpy() == array([1])

    # this not
    df = DataFrame.from_dict({specifier_vbe: [0.5], potential_c: [1]})
    with pytest.raises(KeyError):
        df.ensure_specifier_column(potential_b, debug=True)

    with pytest.raises(KeyError):
        df.ensure_specifier_column(potential_b, debug=True, reference_node="C")


if __name__ == "__main__":
    test_2_voltages_0_potential_and_reference()
    test_1_voltage_1_potential_and_reference()
    test_2_voltage_1_potential_and_reference()
    test_2_voltages_0_potential_and_reference_b_common1()
    test_2_voltages_0_potential_and_reference_b_common2()
    test_2_voltages_0_potential_and_reference_b_common3()



""" DMT interaction with pypi
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import requests
import re
import argparse

try:
    from semver.version import Version as VersionInfo
except ImportError:
    from semver import VersionInfo


def get_file_idx(release, pattern=".whl"):
    return next(file_a for file_a in release if re.search(pattern, file_a["filename"]))


def get_pypi_url(package="DMT-core", version=None, pattern=".whl"):
    package = requests.get(f"https://pypi.python.org/pypi/{package:s}/json").json()
    if version is None:
        version = max(package["releases"].keys())
    # ... check compatibility
    file_to_get = next(
        file for file in package["releases"][version] if re.search(pattern, file["filename"])
    )
    return file_to_get["url"]


def check_version(to_check, package="DMT-core"):
    if to_check.startswith("Version_"):
        to_check = to_check[8:]

    to_check = VersionInfo.parse(to_check)
    if to_check.prerelease is not None and not to_check.prerelease.startswith("rc."):
        raise IOError("The DMT naming schneme is: X.Y.Z-rc.A")
    package = requests.get(f"https://pypi.python.org/pypi/{package:s}/json").json()
    latest_version = max(package["releases"].keys())

    version_pkg = latest_version.split(".")
    if "rc" in version_pkg[2]:
        patch = version_pkg[2].split("rc")[0]
        rc = version_pkg[2][len(patch) + 2 :]
        latest_version = VersionInfo(
            major=version_pkg[0],
            minor=version_pkg[1],
            patch=patch,
            prerelease="rc." + rc,
        )
    else:
        latest_version = VersionInfo(
            major=version_pkg[0], minor=version_pkg[1], patch=version_pkg[2]
        )

    if to_check > latest_version:
        return str(to_check)
    else:
        raise IOError("Version is smaller than or equal as already published version")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Obtain information about a module on pypi.org")
    parser.add_argument(
        "--wheel_link",
        action="store_true",
        help="Pass to obtain a wheel link.",
        required=False,
    )
    parser.add_argument(
        "--check_version",
        type=str,
        help="Check if given version is a valid SemVer version and also a valid next version for the given package.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--package",
        type=str,
        help="Package to query (default: %(default)s).",
        default="DMT-core",
        required=False,
    )
    parser.add_argument(
        "--version",
        type=str,
        help="Version of the query. If None, the newest version is used (default: %(default)s).",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--pattern",
        type=str,
        help="Pattern to scan the different files of the release (default: %(default)s).",
        default=".whl",
        required=False,
    )
    args = parser.parse_args()
    # args = parser.parse_args("--check_version 1.7.1-rc.1".split())

    if args.wheel_link:
        file_link = get_pypi_url(package=args.package, version=args.version, pattern=args.pattern)
        print(file_link)
    elif args.check_version is not None:
        version = check_version(args.check_version, package=args.package)
        print(version)
    else:
        raise IOError()



# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import numpy as np
import pandas as pd
from DMT.core import DutView


class DutDummy(DutView):
    """DutDummy is a dummy subclass of DutView that can be used for testing other modules of DMT. It implements some basic functionality that mimics an actual DutView.

    Methods
    -------
    run_simulation()
        run the simulation for this DutView object using a specified sweep on a specified core.

    Attributes
    ----------
    database_dir   :  string
        Path to the DUT's database.

    name     :  string
        Name of the DUT's database.

    type     :  string
        type of the DUT according to the Dut_type class.

    nodes     : [string]
        List of strings that contain the node names of DUT. Default is none. If nodes is None, nodes will be requested from Dut_type class.
    """

    def __init__(self, database_dir, name, dut_type, modelcard, **kwargs):
        """Initialization routine.

        Parameters
        ----------
        database_dir    : string
            This is the directory were the DUT will create its database.
        name      : string
            This is the name that will be assigned to the DUT.
        dut_type  : Dut_type()
            Type of the DUT.
        nodes     : [string]
            List of strings that contain the node names of DUT. Default is none. If nodes is None, nodes will be requested from Dut_type class.
        """
        self.modelcard = modelcard
        super().__init__(database_dir, name, dut_type, **kwargs)
        self.inp_name = "dummy_input.inp"

    def get_start_sim_command(self):
        return "sleep 1"

    def get_hash(self):
        """Return a hash for the DutView's input file.

        Returns
        -------
        hash : float64
            Hash that corresponds to this DUT.
        """
        mcf = self.modelcard.get("mcf")
        mcf = mcf.value
        return str(mcf)

    # def prepare_simulation(self, sweep):
    #    """ Prepare a dummy simulation.
    #    """
    #    pass

    def import_output_data(self, sweep, delete_sim_results=False):
        """Read the output files that have been produced while simulating sweep and attach them to self.db.

        Parameters
        ----------
        sweep  :  :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated for the desired output files.
        delete_sim_results : {False, True}, optional
            If True, the simulation folder is deleted after reading.

        Raises
        ------
        NotImplementedError
            If the Dut is not a simulatable dut.
        IOError
            If the given sweep can not be read.
        """
        mcf = self.modelcard.get("mcf").value
        data = {}
        df_sweep = sweep.create_df()
        data["V_B"] = df_sweep["V_B"].to_numpy()
        data["I_C"] = np.exp(data["V_B"] / (mcf * 25e-3))
        df = pd.DataFrame(data=data)
        self.manager.save_df(df, self.database_dir / "iv.p")
        return df

    def make_input(self, sweep):
        """Joins simulation header and with a given sweep and returns it!

        Parameters
        ----------
        sweep : :class:`DMT.core.sweep.Sweep`
            Sweep specification according to the Sweep class.
        """
        # self.sweep = sweep
        return "dummy inp"

    def validate_simulation_successful(self, sweep):
        """Checks if the simulation of the given sweep was successful.

        Parameters
        ----------
        sweep  :  :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated.

        Raises
        ------
        NotImplementedError
            If the Dut is not a simulatable dut.
        SimulationUnsuccessful
            If the simulation output is not valid.
        FileNotFoundError
            If the sim log file does not exist.
        """
        pass



""" DMT supplied build latex command
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import subprocess
import re
from pathlib import Path

from DMT.external.os import cd


def build_tex(full_path_to_file, additional_compiler=None, wait=False, extension=None):
    r"""Builds a tex file

    Parameters
    ----------
    full_path_to_file : str or os.Pathlike
    additional_compiler: (str , [str])
        If a additional tex compiler should be used, supply it here. It will be tried as first compiler.
        The additional compiler must be given in this argument with (<call_of_compiler>, List of compiler arguments).
        For example 'latexmk' looks like: ('latexmk', ['--interaction=nonstopmode', '--shell-escape', '--pdf'])
    wait : Boolean, False
        Wait for the build to complete.
    extension : None, optional
        If None, assumed extension of Tex file to build is "tex", else the given string is used. Example: extension="tikz"
    """
    if not isinstance(full_path_to_file, Path):
        full_path_to_file = Path(full_path_to_file)

    ext = ".tex"
    if extension is not None:
        ext = "." + extension.replace(".", "")

    directory = full_path_to_file.parent
    file_name = full_path_to_file.name
    if file_name.endswith(ext):
        file_name = [file_name]
    else:
        file_name = [file_name + ext]

    compilers = (
        ("latexmk", ["--interaction=nonstopmode", "--shell-escape", "--pdf"]),
        ("lualatex", ["--interaction=nonstopmode", "--shell-escape"]),
        ("pdflatex", ["--interaction=nonstopmode", "--shell-escape"]),
    )
    if additional_compiler is not None and additional_compiler[0] is not None:
        compilers = (additional_compiler,) + compilers

    # turn of testing of all compilers. Just use the first one
    compiler = compilers[0]
    command = [compiler[0]] + compiler[1] + file_name

    with cd(directory):
        try:
            if wait:
                # output = subprocess.call(command, stderr=subprocess.STDOUT)
                subprocess.call(command, stderr=subprocess.STDOUT)
            else:
                # output = subprocess.Popen(command, stderr=subprocess.STDOUT)
                subprocess.Popen(command, stderr=subprocess.STDOUT)
        except FileNotFoundError as err:
            print(err)

            # Notify user that the compiler is not found.
            print(
                "The used latex compiler was not found\n"
                + "The latex command was "
                + " ".join(command)
                + "\n"
                + "Make sure that a LaTeX distribution (like Tex Live) is installed and in your PATH."
            )
        except subprocess.CalledProcessError as err:
            # For all other errors print the output and raise the error
            print(err)
            raise
        else:
            print("LaTeX compile successfull")

        # Compilation has finished


def build_png(full_path_to_file, wait=True):
    r"""Builds a tex file to png graphic.

    Parameters
    ----------
    full_path_to_file : str or os.Pathlike

    TODO
    ----
    somehow this does not work so nice...
    """
    if not isinstance(full_path_to_file, Path):
        full_path_to_file = Path(full_path_to_file)
    directory = full_path_to_file.parent
    file_name = full_path_to_file.name.replace(".tex", "")

    with cd(directory):
        # call latex to compile dvi
        command = ["latex", "--interaction=nonstopmode", file_name + ".tex"]
        try:
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        except (OSError, IOError, subprocess.CalledProcessError) as e:
            # For all these errors print the output and raise the error
            print(e.output.decode())
            # raise
        else:
            print(output.decode())

        # call dvisvgm to compile dvi to svg
        command = ["pdftoppm", file_name + ".pdf", "outpuname", "-png", file_name + ".png"]
        # pdftoppm input.pdf outputname -png
        # command = ['convert ', file_name+'.dvi','-quality 90', file_name+'.png']
        # convert -density 300 F_TJ_C.pdf -quality 90 F_TJ_C.png
        try:
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        except (OSError, IOError, subprocess.CalledProcessError) as e:
            # For all these errors print the output and raise the error
            print(e.output.decode())
            # raise
        else:
            print(output.decode())


def build_svg(full_path_to_file, wait=True):
    r"""Builds a tex file to scalable vector graphic.

    Parameters
    ----------
    full_path_to_file : str or os.Pathlike
    """
    if not isinstance(full_path_to_file, Path):
        full_path_to_file = Path(full_path_to_file)
    directory = full_path_to_file.parent
    file_name = full_path_to_file.name.replace(".tex", "")

    with cd(directory):
        # call latex to compile dvi
        command = ["latex", "--interaction=nonstopmode", file_name + ".tex"]
        try:
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        except (OSError, IOError, subprocess.CalledProcessError) as e:
            # For all these errors print the output and raise the error
            print(e.output.decode())
            # raise
        else:
            print(output.decode())

        # call dvisvgm to compile dvi to svg
        command = ["dvisvgm", "--no-fonts", file_name]
        try:
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        except (OSError, IOError, subprocess.CalledProcessError) as e:
            # For all these errors print the output and raise the error
            print(e.output.decode())
            # raise
        else:
            print(output.decode())


def clean_tex_files(directory, file_name, keep=(".tex", ".pdf"), regex=False):
    """Deletes all intermediate files in the directory with name file_name except the tex and pdf.

    Parameters
    ----------
    directory : str
        Direcory to clean
    file_name : str
        File name without file ending (!) to clean
    keep : iterable, optional
        File endings to keep
    """

    def test_str(file_name_to_del, test_file_name):
        return file_name_to_del in test_file_name.stem

    def test_regex(regex_pattern_to_del, test_file_name):
        return bool(re.search(regex_pattern_to_del, test_file_name.stem))

    if regex:
        test_fct = test_regex
    else:
        test_fct = test_str

    for file_curr in Path(directory).iterdir():
        if (
            file_curr.is_file()
            and test_fct(file_name, file_curr)
            and (file_curr.suffix not in keep)
        ):
            file_curr.unlink()


SI_UNITS_CONVERTER = {
    r"\second": "s",
    r"\hertz": "Hz",
    r"\volt": "V",
    r"\watt": "W",
    r"\ampere": "A",
    r"\kelvin": "K",
    r"\ohm": "Ohm",
    r"\siemens": "S",
    r"\farad": "F",
    r"\coulomb": "C",
    r"\electronvolt": "eV",
    r"\meter": "m",
    r"\metre": "m",
    r"\per": "inv",
    r"\square": "sq",
    r"\cubic": "cu",
    r"\giga": "G",
    r"\centi": "c",
    r"\milli": "m",
    r"\micro": "u",
    r"\nano": "n",
    r"\femto": "f",
    r"\pico": "p",
}


def resolve_siunitx(label):
    """This function tries to remove siunitx expressions from given Tex expression."""
    regex_si = r"\\si({.*?})"
    regex_SI = r"\\SI({.*?})({.*?})"
    if "si" in label:
        pattern = re.compile(regex_si)
        for match in pattern.findall(label):
            units = match[1:-1]
            # todo: resolve unit after unit, then for each unit the scaler (milli and so on), than stuff like per. Only then per can be correctly replaced.
            for key in SI_UNITS_CONVERTER:
                units = units.replace(key, SI_UNITS_CONVERTER[key])

            label = label.replace(match, units)

    if "SI" in label:
        pattern = re.compile(regex_SI)
        for match_number, match_unit in pattern.findall(label):
            # match 0 is number
            number = match_number[1:-1]
            label = label.replace(match_number, number)

            # match 1 is unit
            units = match_unit[1:-1]
            for key in SI_UNITS_CONVERTER:
                units = units.replace(key, SI_UNITS_CONVERTER[key])

            label = label.replace(match_unit, units)

    if "underline" in label:
        regex_underline = r"(\\underline{)([^{}]+)(})"
        pattern = re.compile(regex_underline)
        for match in pattern.findall(label):
            label = label.replace(match[0], "")
            label = label.replace(match[1] + match[2], match[1])

    # drop some tex things that are not understood by pandoc
    label = label.replace("\\si", "")
    label = label.replace("\\SI", "")
    label = label.replace("\\,", " ")

    return label


def tex_to_text(tex):
    """Return a text representation of a tex label."""
    tex = resolve_siunitx(tex)
    tex = tex.replace("\\num", "")
    tex = tex.replace("\\mathrm", "")
    tex = tex.replace("\\quad", "  ")
    tex = tex.replace("\\left(", "(")
    tex = tex.replace("\\right)", ")")
    tex = tex.replace("{", "")
    tex = tex.replace("}", "")
    tex = tex.replace("$", "")
    return tex



import logging
from pathlib import Path
from DMT.core import DutMeas, DutType, Plot, DutLib

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent
# -->Start main function
# --->Setup for log
logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_manager.log",
    filemode="w",
)


def test_lib_import():
    """Test if the types are the correct subtypes."""
    # -->Define subroutine at first
    # --->Subroutine: filter_dut
    def filter_dut(dut_name):
        if "dummies" in dut_name:
            return None
        elif "TLM" in dut_name:
            return None
        elif dut_name == "":
            return None
        else:
            dut_transistor = DutMeas(
                database_dir=test_path / "tmp",
                dut_type=DutType.npn,
                force=True,
                wafer=96,
                die="y",
                width=float(0.25e-6),
                length=float(0.25e-6),
                name="dut_meas_npn",
                reference_node="E",
            )

            return dut_transistor

    # --->Arrange DMT class to dmt
    lib = DutLib(
        save_dir=test_path / "tmp",
        force=True,
        AC_filter_names=[("freq_vbc", "ac")],
        DC_filter_names=[("fgummel", "dc")],
    )
    # --->Add source measurement information in dmt and to duts
    lib.import_directory(
        import_dir=folder_path / "test_data" / "0p25x10x1_full",
        dut_filter=filter_dut,
        dut_level=0,
        force=True,
    )
    lib.dut_ref = lib.duts[0]

    # %%
    # --->Add the open and short dummies to the lib
    dut_short = DutMeas(
        database_dir=test_path / "tmp",
        name="dut_short_npn",
        dut_type=DutType.deem_short_bjt,
        reference_node="E",
        force=True,
        wafer=96,
        die="y",
        width=float(0.25e-6),
        length=float(0.25e-6),
    )
    dut_short.add_data(folder_path / "test_data" / "dummy_short_freq.mdm", key="ac")
    dut_short.add_data(folder_path / "test_data" / "short_dc.mdm", key="dc")
    dut_open = DutMeas(
        database_dir=test_path / "tmp",
        name="dut_open_npn",
        dut_type=DutType.deem_open_bjt,
        reference_node="E",
        force=True,
        wafer=96,
        die="y",
        width=float(0.25e-6),
        length=float(0.25e-6),
    )
    dut_open.add_data(folder_path / "test_data" / "dummy_open_freq.mdm", key="ac")
    lib.add_duts([dut_short, dut_open])

    # --->Clean the names of all dataframes, e.g. VB=>V_B
    for dut_a in lib:
        dut_a.clean_data(
            fallback={
                "S_DEEMB(1,1)": None,
                "S_DEEMB(2,1)": None,
                "S_DEEMB(1,2)": None,
                "S_DEEMB(2,2)": None,
            }
        )

    # --->Deembedding only when ac_filter is returned as True
    lib.deembed_AC(False, False, False)
    mres = lib.deembed_DC(False, False, False, t_ref=298)

    print(mres)

    return lib


def test_lib_import_lvl1():
    """Test if the types are the correct subtypes."""
    # -->Define subroutine at first
    # --->Subroutine: filter_dut
    def filter_dut(dut_name):
        if "dummies" in dut_name:
            return None
        elif "TLM" in dut_name:
            return None
        elif dut_name == "":
            return None
        else:
            dut_transistor = DutMeas(
                database_dir=test_path / "tmp",
                dut_type=DutType.npn,
                force=True,
                wafer=96,
                die="y",
                width=float(0.25e-6),
                length=float(0.25e-6),
                name="dut_meas_npn",
                reference_node="E",
            )

            return dut_transistor

    # --->Arrange DMT class to dmt
    lib = DutLib(
        save_dir=test_path / "tmp",
        force=True,
        AC_filter_names=[("freq_vbc", "ac")],
        DC_filter_names=[("fgummel", "dc")],
    )
    # --->Add source measurement information in dmt and to duts
    lib.import_directory(
        import_dir=folder_path / "test_data",
        dut_filter=filter_dut,
        dut_level=1,
        force=True,
    )
    lib.dut_ref = lib.duts[0]

    # %%
    # --->Add the open and short dummies to the lib
    dut_short = DutMeas(
        database_dir=test_path / "tmp",
        name="dut_short_npn",
        dut_type=DutType.deem_short_bjt,
        reference_node="E",
        force=True,
        wafer=96,
        die="y",
        width=float(0.25e-6),
        length=float(0.25e-6),
    )
    dut_short.add_data(folder_path / "test_data" / "dummy_short_freq.mdm", key="ac")
    dut_short.add_data(folder_path / "test_data" / "short_dc.mdm", key="dc")
    dut_open = DutMeas(
        database_dir=test_path / "tmp",
        name="dut_open_npn",
        dut_type=DutType.deem_open_bjt,
        reference_node="E",
        force=True,
        wafer=96,
        die="y",
        width=float(0.25e-6),
        length=float(0.25e-6),
    )
    dut_open.add_data(folder_path / "test_data" / "dummy_open_freq.mdm", key="ac")
    lib.add_duts([dut_short, dut_open])

    # --->Clean the names of all dataframes, e.g. VB=>V_B
    for dut_a in lib:
        dut_a.clean_data(
            fallback={
                "S_DEEMB(1,1)": None,
                "S_DEEMB(2,1)": None,
                "S_DEEMB(1,2)": None,
                "S_DEEMB(2,2)": None,
            }
        )

    # --->Deembedding only when ac_filter is returned as True
    lib.deembed_AC(False, False, False)
    mres = lib.deembed_DC(False, False, False, t_ref=298)

    print(mres)

    return lib


if __name__ == "__main__":
    lib_test = test_lib_import()
    lib_test = test_lib_import_lvl1()
    lib_test.save()

    # --->Read freq related data with "298" and "freq_vbc" from duts
    ft_dfs = []
    for dut in lib_test:
        for key in dut.data.keys():
            if "298" in key and "freq_vbc" in key:
                ft_dfs.append(dut.get_data(key=key))

    # --->Extract ic, ft, vbc from ft_dfs
    ic, ft, vbc, ic_ftmax = [], [], [], []
    for df in ft_dfs:
        df = df.calc_ft("B", "C")
        df = df[df["FREQ"] == 10e9]
        ic.append(df["I_C"])
        ft.append(df["F_T"])
        vbc.append(df["V_B"].to_numpy() - df["V_C"].to_numpy())
        df = df.loc[df["F_T"].idxmax()]
        ic_ftmax.append(df["I_C"])
        vbc[-1] = round(vbc[-1][0], 3)

    # -->Plot
    # --->Define plot type according to Plot
    plt_ft = Plot("f_T(I_C)", x_log=True, style="color", legend_location="upper left")

    # --->Add x,y data in plot with using enumerate: loop index+elements
    for i, array in enumerate(ic):
        plt_ft.add_data_set(
            ic[i] * 1e3, ft[i] * 1e-9, label=r"$V_{\mathrm{BC}}=" + str(vbc[i]) + r"\,$V"
        )

    # --->Define x and y limit
    plt_ft.x_limits = (1e-1, 1e2)
    plt_ft.y_limits = (0, 325)

    # --->Get plot
    plt_ft.plot_pyqtgraph()



""" Testing Circuit and CircuitElement class.
"""
from DMT.core.mcard import MCard
import pytest
from DMT.core import Circuit, CircuitElement
from DMT.core.circuit import RESISTANCE, CAPACITANCE, INDUCTANCE, CURRENT, VOLTAGE, SHORT


def test_CircuitElement():
    # correct way
    ce_r = CircuitElement(
        RESISTANCE,
        "R_1",
        ("A", "B"),
        parameters=[
            ("R", "10"),
        ],
    )

    assert str(ce_r) == "DMT.CircuitElement:R_1 model R nodes A,B"

    # wrong, type must be a string
    with pytest.raises(TypeError):
        CircuitElement(
            5,
            "R_1",
            ("A", "B"),
            parameters=[
                ("R", "10"),
            ],
        )
    # wrong, 'a' is unknown
    with pytest.raises(IOError):
        CircuitElement(
            "a",
            "R_1",
            ("A", "B"),
            parameters=[
                ("R", "10"),
            ],
        )
    # wrong, name must be a string
    with pytest.raises(TypeError):
        CircuitElement(
            CAPACITANCE,
            5,
            ("A", "B"),
            parameters=[
                ("C", "10"),
            ],
        )

    # wrong, nodes must be a tuple
    with pytest.raises(TypeError):
        CircuitElement(
            INDUCTANCE,
            "R_1",
            "A B",
            parameters=[
                ("R", "10"),
            ],
        )
    # wrong, nodes must be tuple of strings
    with pytest.raises(TypeError):
        CircuitElement(
            CURRENT,
            "R_1",
            ("A", 5),
            parameters=[
                ("R", "10"),
            ],
        )

    # wrong, parameters must be list
    with pytest.raises(TypeError):
        CircuitElement(VOLTAGE, "R_1", ("A", "B"), parameters="R=10")
    # wrong, parameters must be list of tuples
    with pytest.raises(TypeError):
        CircuitElement(RESISTANCE, "R_1", ("A", "B"), parameters=["R=10"])
    # wrong, parameters must be list of tuples of strings
    with pytest.raises(TypeError):
        CircuitElement(RESISTANCE, "R_1", ("A", "B"), parameters=[("R", 10)])

    # correct way
    CircuitElement(SHORT, "I_S", ("A", "B"))


def test_CircuitElement_MCard():
    # mcards
    mcard = MCard(["A", "D"], default_subckt_name="dio", default_module_name="diova")

    element = CircuitElement(
        mcard.default_module_name, mcard.default_subckt_name, ["n_A", "n_D"], parameters=mcard
    )

    assert element.parameters == mcard


def test_circuit():
    # correct
    Circuit(
        [
            CircuitElement(
                RESISTANCE,
                "R_1",
                ("A", "B"),
                parameters=[
                    ("R", "10"),
                ],
            ),
            CircuitElement(SHORT, "I_S", ("A", "B")),
            "V_A=0",
        ]
    )
    # wrong
    with pytest.raises(TypeError):
        Circuit(
            [
                CircuitElement(
                    RESISTANCE,
                    "R_1",
                    ("A", "B"),
                    parameters=[
                        ("R", "10"),
                    ],
                ),
                CircuitElement(SHORT, "I_S", ("A", "B")),
                "V_A=0",
                5,
            ]
        )

    # outdated and now forbidden
    with pytest.raises(NotImplementedError):
        Circuit("common_emitter")

    # also test the default circuit as this has to be implemented in the subclass
    mcard = MCard(["A", "D"], default_subckt_name="dio", default_module_name="diova")
    with pytest.raises(NotImplementedError):
        mcard.get_circuit()  # should always work...


if __name__ == "__main__":
    test_CircuitElement()
    test_CircuitElement_MCard()
    test_circuit()
